--
-- Preamble for non-privileged portion of database
--


\c pscheduler

-- TODO: This should be taken care of by postgresql-load but isn't.
SET ROLE pscheduler;


-- Drop connected applications to prevent getting stuck behind
-- long-running queries started by applications that were killed
-- without the server noticing.

DO $$
BEGIN
    PERFORM pg_terminate_backend(pg_stat_activity.pid)
    FROM pg_stat_activity
    WHERE
        pg_stat_activity.datname = 'pscheduler'
        AND usename = 'pscheduler'
        AND pid <> pg_backend_pid()     -- Not this connection
        AND application_name <> 'psql'  -- Not interactive sessions
        ;
END;
$$ LANGUAGE plpgsql;



BEGIN TRANSACTION;
--
-- UTILITIES
-- 
--


-- ----------------------------------------------------------------------------

--
-- CONVERSIONS
--

-- Most of these functions take a TEXT value and attempt to convert it
-- into the return type.  If the 'picky' argument is TRUE, the
-- function will raise an exception if the conversion fails, otherwise
-- it will return NULL.

-- TODO: In all of these, need to look at whether we should narrow the
-- exception down to something more specific than 'OTHERS'.

DO $$ BEGIN PERFORM drop_function_all('text_to_timestamp_with_time_zone'); END $$;

CREATE OR REPLACE FUNCTION 
text_to_timestamp_with_time_zone(
	value TEXT,
	picky BOOL DEFAULT TRUE
)
RETURNS TIMESTAMP WITH TIME ZONE
AS $$
DECLARE
	converted TIMESTAMP WITH TIME ZONE DEFAULT NULL;
BEGIN

    -- Special case: "forever" and "infinity" map to Pg's infinity.
    -- Regular SQL would have to use the largest value.
    IF value = 'forever' THEN
       RETURN 'infinity';
    END IF;

    BEGIN
	converted := value::TIMESTAMP WITH TIME ZONE;
    EXCEPTION WHEN OTHERS THEN
        IF picky THEN
	    RAISE EXCEPTION 'Invalid timestamp "%"', value;
	ELSE
	    RETURN NULL;
	END IF;	
    END;
RETURN converted;
END;
$$ LANGUAGE plpgsql;



DO $$ BEGIN PERFORM drop_function_all('timestamp_with_time_zone_to_iso8601'); END $$;

CREATE OR REPLACE FUNCTION 
timestamp_with_time_zone_to_iso8601(
	value TIMESTAMP WITH TIME ZONE
)
RETURNS TEXT
AS $$
BEGIN

    IF value IS NULL
    THEN
        RETURN NULL;
    END IF;

    return to_json(date_trunc('seconds', value)) #>> '{}';

END;
$$ LANGUAGE plpgsql;


DO $$ BEGIN PERFORM drop_function_all('interval_to_iso8601'); END $$;

CREATE OR REPLACE FUNCTION interval_to_iso8601(period INTERVAL)
RETURNS TEXT
AS $$
DECLARE
    extracted NUMERIC;
    days NUMERIC;
    weeks NUMERIC;
    use_t BOOLEAN;
    iso TEXT;
BEGIN

    iso = '';
    use_t = FALSE;

    extracted = EXTRACT('SECONDS' FROM period);
    IF extracted > 0
    THEN
        iso := to_char(extracted, 'FM999999999') || 'S' || iso;
        use_t := TRUE;
    END IF;

    extracted = EXTRACT('MINUTES' FROM period);
    IF extracted > 0
    THEN
        iso := to_char(extracted, 'FM999999999') || 'M' || iso;
        use_t := TRUE;
    END IF;

    extracted = EXTRACT('HOURS' FROM period);
    IF extracted > 0
    THEN
        iso := to_char(extracted, 'FM999999999') || 'H' || iso;
        use_t := TRUE;
    END IF;

    IF use_t
    THEN
        iso := 'T' || iso;
    END IF;

    extracted = EXTRACT('DAYS' FROM period);
    IF extracted > 0
    THEN

        days := extracted % 7;
	IF days > 0
        THEN
            iso := to_char(days, 'FM999999999') || 'D' || iso;
        END IF;

        extracted := extracted - days;
        IF extracted > 0
        THEN
            iso := to_char(extracted / 7, 'FM999999999') || 'W' || iso;
        END IF;

    END IF;

    IF ISO = ''
    THEN
        RETURN 'P0D';
    END IF;

    RETURN 'P' || iso;
END;
$$ LANGUAGE plpgsql;




DO $$ BEGIN PERFORM drop_function_all('text_to_interval'); END $$;

CREATE OR REPLACE FUNCTION 
text_to_interval(
	value TEXT,
	picky BOOL DEFAULT TRUE
)
RETURNS INTERVAL
AS $$
DECLARE
	converted INTERVAL DEFAULT NULL;
BEGIN
    BEGIN
	converted := value::INTERVAL;
    EXCEPTION WHEN OTHERS THEN
        IF picky THEN
	    RAISE EXCEPTION 'Invalid interval "%"', value;
	ELSE
	    RETURN NULL;
	END IF;	
    END;
RETURN converted;
END;
$$ LANGUAGE plpgsql;



DO $$ BEGIN PERFORM drop_function_all('text_to_numeric'); END $$;

CREATE OR REPLACE FUNCTION 
text_to_numeric(
	value TEXT,
	picky BOOL DEFAULT TRUE
)
RETURNS NUMERIC
AS $$
DECLARE
	converted NUMERIC DEFAULT NULL;
BEGIN
    BEGIN
	converted := value::NUMERIC;
    EXCEPTION WHEN OTHERS THEN
        IF picky THEN
	    RAISE EXCEPTION 'Invalid numeric "%"', value;
	ELSE
	    RETURN NULL;
	END IF;	
    END;
RETURN converted;
END;
$$ LANGUAGE plpgsql;


DO $$ BEGIN PERFORM drop_function_all('text_to_jsonb'); END $$;

CREATE OR REPLACE FUNCTION 
text_to_jsonb(
	value TEXT,
	picky BOOL DEFAULT TRUE
)
RETURNS JSONB
AS $$
DECLARE
	converted JSONB DEFAULT NULL;
BEGIN
    BEGIN
	converted := value::JSONB;
    EXCEPTION WHEN OTHERS THEN
        IF picky THEN
	    RAISE EXCEPTION 'Invalid JSON "%"', value;
	ELSE
	    RETURN NULL;
	END IF;	
    END;
RETURN converted;
END;
$$ LANGUAGE plpgsql;


-- ----------------------------------------------------------------------------

--
-- DATE AND TIME
--

-- Truncate a timestamp to our scheduling increment (seconds)

DO $$ BEGIN PERFORM drop_function_all('normalized_time'); END $$;

CREATE OR REPLACE FUNCTION normalized_time(value TIMESTAMP WITH TIME ZONE)
RETURNS TIMESTAMP WITH TIME ZONE
AS $$
BEGIN
	RETURN date_trunc('seconds', value);
END;
$$LANGUAGE plpgsql;


-- Return the normalized current (transaction start) time

DO $$ BEGIN PERFORM drop_function_all('normalized_now'); END $$;

CREATE OR REPLACE FUNCTION normalized_now()
RETURNS TIMESTAMP WITH TIME ZONE
AS $$
BEGIN
	RETURN normalized_time(now());
END;
$$LANGUAGE plpgsql;


-- Return the normalized wall clock time

DO $$ BEGIN PERFORM drop_function_all('normalized_wall_clock'); END $$;

CREATE OR REPLACE FUNCTION normalized_wall_clock()
RETURNS TIMESTAMP WITH TIME ZONE
AS $$
BEGIN
	RETURN normalized_time(clock_timestamp());
END;
$$LANGUAGE plpgsql;




-- Practical alternatives to infinite timestamps for use with intervals.
--
-- These exist because PostgreSQL does not yet support infinite
-- intervals (been in the plans since before 2008, apparently).  As a
-- substitute, the minimum/maximum dates supported by the TIMESTAMP
-- WITH TIME ZONE type are provided here.
--
-- See:
--   https://wiki.postgresql.org/wiki/Todo#Dates_and_Times
--   http://www.postgresql.org/docs/9.5/static/datatype-datetime.html


DO $$ BEGIN PERFORM drop_function_all('tstz_negative_infinity'); END $$;

CREATE OR REPLACE FUNCTION tstz_negative_infinity()
RETURNS TIMESTAMP WITH TIME ZONE
AS $$
BEGIN
	RETURN '4713-01-01 BC'::TIMESTAMP WITH TIME ZONE;
END;
$$LANGUAGE plpgsql;


DO $$ BEGIN PERFORM drop_function_all('tstz_infinity'); END $$;

CREATE OR REPLACE FUNCTION tstz_infinity()
RETURNS TIMESTAMP WITH TIME ZONE
AS $$
BEGIN
	RETURN '294276-12-31'::TIMESTAMP WITH TIME ZONE;
END;
$$LANGUAGE plpgsql;





-- Find the next even occurence of a specified interval, truncating
-- fractional seconds from both.  For example, '2016-02-02
-- 11:27:36-05' and 'PT1H' would yield '2016-02-02 12:00:00-05'

DO $$ BEGIN PERFORM drop_function_all('time_next_interval'); END $$;

CREATE OR REPLACE FUNCTION time_next_interval(
    start TIMESTAMP WITH TIME ZONE,
    round_to INTERVAL
)
RETURNS TIMESTAMP WITH TIME ZONE
AS $$
DECLARE
    start_epoch NUMERIC;
    round_to_epoch NUMERIC;
BEGIN
    -- Convert times to epoch with fractional seconds

    start_epoch := EXTRACT(EPOCH FROM start);
    round_to_epoch := EXTRACT(EPOCH FROM round_to);

    RETURN to_timestamp(
        (TRUNC(start_epoch / round_to_epoch)::NUMERIC + 1)
	* round_to_epoch
	);
END;
$$ LANGUAGE plpgsql;


-- Round an interval up to the next whole second if there are any
-- fractional seconds.

DO $$ BEGIN PERFORM drop_function_all('interval_round_up'); END $$;

CREATE OR REPLACE function interval_round_up(candidate INTERVAL)
RETURNS INTERVAL
AS $$
DECLARE
    microseconds INTEGER;
BEGIN
    microseconds := EXTRACT(MICROSECONDS FROM candidate)::INTEGER % 1000000;
    IF microseconds > 0 THEN
        microseconds := 1000000 - microseconds;
	candidate = candidate + make_interval(0, 0, 0, 0, 0, 0, microseconds / 1000000.0);
    END IF;

    return CANDIDATE;
END;
$$ LANGUAGE plpgsql;



-- ----------------------------------------------------------------------------

--
-- STRING AND URL
--

-- Source: https://stackoverflow.com/a/33616365/180674

DO $$ BEGIN PERFORM drop_function_all('url_encode'); END $$;

CREATE OR REPLACE FUNCTION uri_encode(input text)
  RETURNS text
  LANGUAGE plpgsql
  IMMUTABLE STRICT
AS $$
DECLARE
  parsed text;
  safePattern text;
BEGIN
  safePattern = 'a-zA-Z0-9_~/\-\.';
  IF input ~ ('[^' || safePattern || ']') THEN
    SELECT STRING_AGG(fragment, '')
    INTO parsed
    FROM (
      SELECT prefix || encoded AS fragment
      FROM (
        SELECT COALESCE(match[1], '') AS prefix,
               COALESCE('%' || encode(match[2]::bytea, 'hex'), '') AS encoded
        FROM (
          SELECT regexp_matches(
            input,
            '([' || safePattern || ']*)([^' || safePattern || '])?',
            'g') AS match
        ) matches
      ) parsed
    ) fragments;
    RETURN parsed;
  ELSE
    RETURN input;
  END IF;
END;
$$
;



-- Generate a random string

DO $$ BEGIN PERFORM drop_function_all('random_string'); END $$;

CREATE OR REPLACE FUNCTION random_string(
  len INTEGER,                         -- String length
  alphanumeric BOOLEAN DEFAULT FALSE,  -- Alphanumeric only
  random_len BOOLEAN DEFAULT FALSE     -- Make length random between len/2 and len
)
  RETURNS text
  LANGUAGE plpgsql
  IMMUTABLE STRICT
AS $$
DECLARE
  charset TEXT;
BEGIN

  charset := 'ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789';
  IF NOT alphanumeric THEN
    charset := CONCAT(charset, '!"#$%&''()*+,-./:;<=>?@[\]^_`{|}~');
  END IF;

  IF random_len THEN
     -- Chop off up to half of the length
     len := (len - (random() * (len / 2)))::integer;
  END IF;

  RETURN array_to_string(array(
    SELECT SUBSTR(charset, ((random()*(LENGTH(charset)-1)+1)::integer), 1)
    FROM generate_series(1,len)), '');

END;
$$
;


-- ----------------------------------------------------------------------------

--
-- DEBUGGING TOOLS
--

DROP VIEW IF EXISTS queries;
CREATE OR REPLACE VIEW queries
AS
    SELECT
        application_name,
        state,
        CASE state
            WHEN 'active' THEN (now() - query_start)::TEXT
            ELSE 'FINISHED'
            END
            AS run_time,
            regexp_replace(query, '[ \t]+', ' ', 'g') AS query
FROM pg_stat_activity
ORDER BY run_time desc
;
--
-- Table that keeps track of the version numbers of other tables and
-- supporting utility functions.
--

-- See the README file in this directory for details.


-- Find the installed version of a table and check for consistency
-- between the database and the versions table.
--
-- Returns 0 if the table does not exist.

DO $$ BEGIN PERFORM drop_function_all('table_version_find'); END $$;

CREATE OR REPLACE FUNCTION table_version_find(
    table_name   TEXT,
    table_schema TEXT DEFAULT 'public'
)
RETURNS INTEGER
AS $$
DECLARE
    table_exists BOOLEAN;
    table_version INTEGER;
BEGIN

    table_exists := EXISTS (
        SELECT *
        FROM pg_catalog.pg_tables
        WHERE
            schemaname = table_schema
            AND tablename = table_name
    );

    -- The table_version table is a special case, because if it
    -- doesn't exist, there's no way it can have a row saying it
    -- exists.

    IF NOT table_exists AND table_name = 'table_version'
    THEN
        RETURN 0;
    END IF;

    -- Find the version of the table.

    SELECT INTO table_version version FROM table_version WHERE name = table_name;
    IF NOT FOUND
    THEN
        table_version := 0;
    END IF;

    -- This is an untenable situation because we would have no idea
    -- what version of the table is installed and couldn't do upgrades
    -- to it.

    IF table_exists AND table_version = 0
    THEN
        RAISE EXCEPTION 'Table "%" exists but is not in the versions table.',
	    table_name;
    END IF;


    RETURN table_version;

END;
$$ LANGUAGE plpgsql;


DO $$ BEGIN PERFORM drop_function_all('table_version_set'); END $$;

CREATE OR REPLACE FUNCTION table_version_set(
    t_name TEXT,
    t_version INTEGER,
    t_version_old INTEGER
)
RETURNS VOID
AS $$
BEGIN

    IF t_version < t_version_old
    THEN
        RAISE EXCEPTION '%s: Cannot upgrade from version % to %',
            t_name, t_version, t_version_old;
     END IF;

    -- Don't do an upsert here, because the trigger depends on seeing
    -- an UPDATE to vary the messages it produces.

    IF t_version_old = 0
    THEN
        INSERT INTO table_version (name, version)
            VALUES (t_name, t_version);
    ELSE
        UPDATE table_version
        SET version = t_version
        WHERE name = t_name;
    END IF;

END;
$$ LANGUAGE plpgsql;




DO $$
DECLARE
    t_name TEXT;            -- Name of the table being worked on
    t_version INTEGER;      -- Current version of the table
    t_version_old INTEGER;  -- Prior version of the table
BEGIN

    -- WARNING: Please see the README file before modifying this code.

    --
    -- Preparation
    --

    t_name := 'table_version';

    t_version := table_version_find(t_name);
    t_version_old := t_version;


    --
    -- Upgrade Blocks
    --

    IF t_version = 0
    THEN

        CREATE TABLE table_version (
	    -- Name of the table
            name     		   TEXT
	    			   UNIQUE
				   NOT NULL,

	    -- Version of the table
            version    		   INTEGER
	    			   NOT NULL
                                   CHECK (version > 0),

	    -- Last attempt at an upgrade
            attempted              TIMESTAMP WITH TIME ZONE
	    			   DEFAULT now(),

	    -- Last time the table was modified
            modified               TIMESTAMP WITH TIME ZONE
	    			   DEFAULT now()

        );

	t_version := t_version + 1;

    END IF;

    -- -- 1 to 2
    -- IF t_version = 1
    -- THEN
    --     -- ALTER TABLE ...
    --     t_version := t_version + 1;
    -- END IF;

    --
    -- Cleanup
    --

    PERFORM table_version_set(t_name, t_version, t_version_old);

END;
$$ LANGUAGE plpgsql;



DROP TRIGGER IF EXISTS table_version_alter ON table_version CASCADE;

DO $$ BEGIN PERFORM drop_function_all('table_version_alter'); END $$;

CREATE OR REPLACE FUNCTION table_version_alter()
RETURNS TRIGGER
AS $$
BEGIN

    NEW.attempted = now();

    IF TG_OP = 'UPDATE'
    THEN

        -- Downgrades are not acceptable.
        IF NEW.version < OLD.version
        THEN
            RAISE EXCEPTION 'Table %s: Cannot be downgraded (%s to %s)',
    	    NEW.name, OLD.version, NEW.version;
        END IF;

        IF NEW.version > OLD.version
        THEN
            NEW.modified = now();
        END IF;

        IF NEW.version <> OLD.version
        THEN
            RAISE NOTICE 'Table %: Upgraded from version % to %',
                NEW.name, OLD.version, NEW.version;
        END IF;

    ELSIF TG_OP = 'INSERT'
    THEN

        RAISE NOTICE 'Table %: Created at version %', NEW.name, NEW.version;

    END IF;


    RETURN NEW;
END;
$$ LANGUAGE plpgsql;

CREATE TRIGGER table_version_alter
BEFORE INSERT OR UPDATE
ON table_version
FOR EACH ROW
    EXECUTE PROCEDURE table_version_alter();



DROP TRIGGER IF EXISTS table_version_alter_after ON table_version CASCADE;

DO $$ BEGIN PERFORM drop_function_all('table_version_alter_after'); END $$;

CREATE OR REPLACE FUNCTION table_version_alter_after()
RETURNS TRIGGER
AS $$
BEGIN

    -- Remove tables that have been dropped from the database.

    DELETE FROM table_version 
    WHERE NOT EXISTS (
        SELECT *
        FROM pg_catalog.pg_tables
        WHERE tablename = table_version.name
    );

    RETURN NEW;
END;
$$ LANGUAGE plpgsql;

CREATE TRIGGER table_version_alter_after
AFTER INSERT OR UPDATE
ON table_version
FOR EACH STATEMENT
    EXECUTE PROCEDURE table_version_alter_after();
--
-- Table with one row that holds configurable values
--


DO $$
DECLARE
    t_name TEXT;            -- Name of the table being worked on
    t_version INTEGER;      -- Current version of the table
    t_version_old INTEGER;  -- Version of the table at the start
BEGIN

    --
    -- Preparation
    --

    t_name := 'configurables';

    t_version := table_version_find(t_name);
    t_version_old := t_version;


    --
    -- Upgrade Blocks
    --

    -- Version 0 (nonexistant) to version 1
    IF t_version = 0
    THEN

        CREATE TABLE configurables (

            -- How far in advance we should schedule runs
            schedule_horizon    INTERVAL
        			DEFAULT 'P1D',

            -- How long we should keep old runs and tasks
            keep_runs_tasks	INTERVAL
        			DEFAULT 'P7D',

            -- Maximum runs in parallel
            max_parallel_runs	INTEGER
        			DEFAULT 50
        );

        -- This table gets exactly one row that can only ever be updated.
        INSERT INTO configurables DEFAULT VALUES;

	t_version := t_version + 1;

    END IF;

    -- Version 1 to version 2
    IF t_version = 1
    THEN
        ALTER TABLE configurables ADD COLUMN
        -- How long runs should be late before being considered stragglers
        run_straggle INTERVAL DEFAULT 'PT10S';

        t_version := t_version + 1;
    END IF;

    -- Version 2 to version 3
    IF t_version = 2
    THEN
        -- Do this update only if the old default is in place.
        UPDATE configurables SET run_straggle = 'PT30S'
        WHERE run_straggle = 'PT10S';

        t_version := t_version + 1;
    END IF;

    -- Version 3 to version 4
    -- Drops unused max_parallel_runs column
    IF t_version = 3
    THEN
        ALTER TABLE configurables DROP COLUMN max_parallel_runs;

        t_version := t_version + 1;
    END IF;

    -- Version 4 to version 5
    -- Changes default keep time to two days.
    IF t_version = 4
    THEN
        -- Make sure pre-version-4 triggers don't fire.
        ALTER TABLE configurables DISABLE TRIGGER USER;

        ALTER TABLE configurables
        ALTER COLUMN keep_runs_tasks
        SET DEFAULT 'P2D';

        UPDATE configurables SET keep_runs_tasks = DEFAULT;

        ALTER TABLE configurables ENABLE TRIGGER USER;

        t_version := t_version + 1;
    END IF;


    -- Version 5 to version 6
    -- Add run start margin
    IF t_version = 5
    THEN

        ALTER TABLE configurables DISABLE  TRIGGER USER;

        ALTER TABLE configurables
        ADD COLUMN
	run_start_margin INTERVAL DEFAULT 'PT3S';

        UPDATE configurables SET run_start_margin = DEFAULT;

        ALTER TABLE configurables ENABLE TRIGGER USER;

        t_version := t_version + 1;
    END IF;


    --
    -- Cleanup
    --

    PERFORM table_version_set(t_name, t_version, t_version_old);

END;
$$ LANGUAGE plpgsql;



DROP TRIGGER IF EXISTS configurables_update ON configurables CASCADE;

DO $$ BEGIN PERFORM drop_function_all('configurables_update'); END $$;

CREATE OR REPLACE FUNCTION configurables_update()
RETURNS TRIGGER
AS $$
BEGIN
    NOTIFY configurables_changed;
    RETURN NEW;
END;
$$ LANGUAGE plpgsql;

CREATE TRIGGER configurables_update
AFTER UPDATE
ON configurables
FOR EACH ROW
    EXECUTE PROCEDURE configurables_update();



DROP TRIGGER IF EXISTS configurables_alter ON configurables CASCADE;
DROP TRIGGER IF EXISTS configurables_truncate ON configurables CASCADE;

DO $$ BEGIN PERFORM drop_function_all('configurables_noalter'); END $$;

CREATE OR REPLACE FUNCTION configurables_noalter()
RETURNS TRIGGER
AS $$
BEGIN
	RAISE EXCEPTION 'This table can only be updated.';
END;
$$ LANGUAGE plpgsql;

CREATE TRIGGER configurables_alter
BEFORE INSERT OR DELETE
ON configurables
FOR EACH ROW
    EXECUTE PROCEDURE configurables_noalter();

CREATE TRIGGER configurables_truncate
BEFORE TRUNCATE
ON configurables
EXECUTE PROCEDURE configurables_noalter();


--
-- Maintenance
--

-- Maintenance functions

DO $$ BEGIN PERFORM drop_function_all('configurables_maint_minute'); END $$;

CREATE OR REPLACE FUNCTION configurables_maint_minute()
RETURNS VOID
AS $$
DECLARE
    command_result external_program_result;
    config_json JSONB;
BEGIN

    -- Re-read the configuration

    command_result := pscheduler_command(ARRAY['validate-configurables', '--dump']);
    IF command_result.status <> 0 THEN
        -- TODO: Decide whether we want to log this once a minute.
        -- RAISE NOTICE 'Unable to read configurables: %', command_result.stderr;
        RETURN;
    END IF;

    config_json := command_result.stdout::JSONB;
    IF config_json IS NULL THEN
        RETURN;
    END IF;

    -- Update the database

    IF config_json->'keep-runs-tasks' IS NOT NULL THEN
        UPDATE configurables SET keep_runs_tasks = (config_json ->> 'keep-runs-tasks')::INTERVAL;
    END IF;

    IF config_json->'run-straggle' IS NOT NULL THEN
        UPDATE configurables SET run_straggle = (config_json ->> 'run-straggle')::INTERVAL;
    END IF;

END;
$$ LANGUAGE plpgsql;
--
-- Table of daemon heartbeat records
--

DO $$
DECLARE
    t_name TEXT;            -- Name of the table being worked on
    t_version INTEGER;      -- Current version of the table
    t_version_old INTEGER;  -- Version of the table at the start
BEGIN

    --
    -- Preparation
    --

    t_name := 'heartbeat';

    t_version := table_version_find(t_name);
    t_version_old := t_version;


    --
    -- Upgrade Blocks
    --

    -- Version 0 (nonexistant) to version 1
    IF t_version = 0
    THEN

        CREATE TABLE heartbeat (

        	-- Row identifier
        	name		TEXT
        			PRIMARY KEY,
        	-- Start time
        	started		TIMESTAMP WITH TIME ZONE,

        	-- Update count
        	updates		BIGINT DEFAULT 0,

        	-- When this record was last updated
        	last		TIMESTAMP WITH TIME ZONE,

        	-- How long until the next heartbeat is expected, NULL if we don't know.
        	next_time	INTERVAL
        );

	t_version := t_version + 1;

    END IF;

    -- Version 1 to version 2
    -- 
    IF t_version = 1
    THEN
        ALTER TABLE heartbeat
        ALTER COLUMN started SET DEFAULT now();
    
        t_version := t_version + 1;
    END IF;


    --
    -- Cleanup
    --

    PERFORM table_version_set(t_name, t_version, t_version_old);

END;
$$ LANGUAGE plpgsql;



-- This was in an older version.

DROP TRIGGER IF EXISTS heartbeat_alter ON heartbeat CASCADE;
DO $$ BEGIN PERFORM drop_function_all('heartbeat_alter'); END $$;
 


-- Start a fresh heartbeat record for a program

DO $$ BEGIN PERFORM drop_function_all('heartbeat_boot'); END $$;

CREATE OR REPLACE FUNCTION heartbeat_boot(new_name TEXT)
RETURNS VOID
AS $$
BEGIN

    INSERT INTO heartbeat (name, last) VALUES (new_name, now())
    ON CONFLICT (name) DO UPDATE
    SET started = now(), updates = 0, last = now();

END;
$$ LANGUAGE plpgsql;


-- Insert a new heartbeat or update an existing one by name

DO $$ BEGIN PERFORM drop_function_all('heartbeat'); END $$;

CREATE OR REPLACE FUNCTION heartbeat(new_name TEXT, new_next_time INTERVAL DEFAULT NULL)
RETURNS VOID
AS $$
BEGIN

    INSERT INTO heartbeat (name, next_time, updates, last)
    VALUES (new_name, new_next_time, 0, now())
    ON CONFLICT (name) DO UPDATE
    SET
        next_time = new_next_time,
        updates = heartbeat.updates + 1,
	last = now()
    ;

END;
$$ LANGUAGE plpgsql;



-- Function to run at startup.

DO $$ BEGIN PERFORM drop_function_all('heartbeat_cold_boot'); END $$;

CREATE OR REPLACE FUNCTION heartbeat_cold_boot()
RETURNS VOID
AS $$
BEGIN
    DELETE FROM heartbeat;
    INSERT INTO heartbeat (name)
    VALUES
        ('ticker'),
	('scheduler'),
	('runner'),
	('archiver')
    ;
END;
$$ LANGUAGE plpgsql;



-- Full-content view

DROP VIEW IF EXISTS heartbeat_full CASCADE;

CREATE OR REPLACE VIEW heartbeat_full
AS
    SELECT
        name,
	interval_to_iso8601(now() - started) AS uptime,
	updates,
        timestamp_with_time_zone_to_iso8601(last) AS last,
	CASE
	    WHEN next_time IS NOT NULL
	        THEN interval_to_iso8601(next_time)
	    ELSE NULL
	END AS next_time,
	CASE
	    WHEN next_time IS NOT NULL
	        THEN timestamp_with_time_zone_to_iso8601(last + next_time)
	    ELSE NULL
	END AS expected,
        CASE
            WHEN (last + next_time) < now()
	        THEN interval_to_iso8601(now() - (last + next_time))
            ELSE NULL
        END AS overdue
    FROM heartbeat
    ORDER BY NAME
;


-- Fully-aggregated JSON for the API to pass along to curious clients.

CREATE OR REPLACE VIEW heartbeat_json
AS
    SELECT json_object_agg(service, status) AS heartbeat_json
    FROM
        ( SELECT
              name AS service,
              ( SELECT (json_agg(service_tbl)->0)::JSONB - 'name'
                FROM (SELECT * FROM heartbeat_full hf where hf.name = heartbeat_full.name) service_tbl
              ) AS status
          FROM heartbeat_full
        ) status_table
;
--
-- Table of Authorization Key Types
--

DO $$
DECLARE
    t_name TEXT;            -- Name of the table being worked on
    t_version INTEGER;      -- Current version of the table
    t_version_old INTEGER;  -- Version of the table at the start
BEGIN

    --
    -- Preparation
    --

    t_name := 'auth_key_type';

    t_version := table_version_find(t_name);
    t_version_old := t_version;


    --
    -- Upgrade Blocks
    --

    -- Version 0 (nonexistant) to version 1
    IF t_version = 0
    THEN

        CREATE TABLE auth_key_type (

        	-- Row identifier
        	id		BIGSERIAL
        			PRIMARY KEY,

        	-- Type Name
        	name		TEXT
        			UNIQUE NOT NULL
        );

	t_version := t_version + 1;

    END IF;

    -- Version 1 to version 2
    -- IF t_version = 1
    -- THEN
    --     ,,,
    -- 
    --     t_version := t_version + 1;
    -- END IF;


    --
    -- Cleanup
    --

    PERFORM table_version_set(t_name, t_version, t_version_old);

END;
$$ LANGUAGE plpgsql;
--
-- Table of Authorization Keys
--


DO $$
DECLARE
    t_name TEXT;            -- Name of the table being worked on
    t_version INTEGER;      -- Current version of the table
    t_version_old INTEGER;  -- Version of the table at the start
BEGIN

    --
    -- Preparation
    --

    t_name := 'auth_key';

    t_version := table_version_find(t_name);
    t_version_old := t_version;


    --
    -- Upgrade Blocks
    --

    -- Version 0 (nonexistant) to version 1
    IF t_version = 0
    THEN

        CREATE TABLE auth_key (

        	-- Row identifier
        	id		BIGSERIAL
        			PRIMARY KEY,

        	-- Key type
        	type		BIGINT
        			REFERENCES auth_key_type(id)
        			ON DELETE CASCADE,

        	-- User name
        	name		TEXT
        			NOT NULL,

        	-- Hashed password
        	password	TEXT
        			NOT NULL,

        	-- When this record was last updated
        	updated		TIMESTAMP WITH TIME ZONE
        			DEFAULT now(),

        	UNIQUE		(type, name)

        );


	-- There's no sense in indexing the password because the salt
	-- forces a computation against it.  This should get it down
	-- to one row.

        CREATE INDEX auth_key_type_name ON auth_key(type, name);

	t_version := t_version + 1;

    END IF;

    -- Version 1 to version 2
    -- Remove unused and trouble-causing version column
    -- IF t_version = 1
    -- THEN
    --     ...
    -- 
    --     t_version := t_version + 1;
    -- END IF;


    --
    -- Cleanup
    --

    PERFORM table_version_set(t_name, t_version, t_version_old);

END;
$$ LANGUAGE plpgsql;



-- Insert a new key or update an existing one by name

DO $$ BEGIN PERFORM drop_function_all('auth_key_add_update'); END $$;

CREATE OR REPLACE FUNCTION auth_key_add_update(type TEXT, key_name TEXT, password TEXT)
RETURNS VOID
AS $$
DECLARE
    type_id BIGINT;
    crypted TEXT;
BEGIN

    SELECT INTO type_id id FROM auth_key_type WHERE auth_key_type.name = type;
    IF NOT FOUND THEN
        RAISE EXCEPTION 'Unknown key type "%"', type;
    END IF;

    -- Note that this uses MD5 because it takes infinitely-long passwords
    -- TODO: Consider truncating to 72 and using bf?
    crypted := crypt(password, gen_salt('md5'));

    -- This can't be done with ON CONFLICT because multiple columns
    -- cause problems.  Use an old-fashioned UPSERT.

    IF EXISTS (SELECT * FROM auth_key WHERE auth_key.type = type_id AND name = key_name)
    THEN
        UPDATE auth_key
        SET password = crypted
	WHERE auth_key.type = type_id AND name = key_name;
    ELSE
        INSERT INTO auth_key (type, name, password)
        VALUES (type_id, key_name, crypted);
    END IF;

END;
$$ LANGUAGE plpgsql;



-- See if a provided type/name/password are valid

DO $$ BEGIN PERFORM drop_function_all('auth_key_is_valid'); END $$;

CREATE OR REPLACE FUNCTION auth_key_is_valid(type TEXT, key_name TEXT, pass TEXT)
RETURNS BOOLEAN
AS $$
DECLARE
    type_id BIGINT;
BEGIN

    SELECT INTO type_id id FROM auth_key_type WHERE auth_key_type.name = type;
    IF NOT FOUND THEN
        RAISE EXCEPTION 'Unknown key type "%"', type;
    END IF;

    RETURN EXISTS (
        SELECT * FROM auth_key
        WHERE
	    auth_key.type = type_id
	    AND name = key_name
	    AND password = crypt(pass, password)
	);

END;
$$ LANGUAGE plpgsql;



-- Delete a key

DO $$ BEGIN PERFORM drop_function_all('auth_key_delete'); END $$;

CREATE OR REPLACE FUNCTION auth_key_delete(type TEXT, key_name TEXT)
RETURNS VOID
AS $$
DECLARE
    type_id BIGINT;
    crypted TEXT;
BEGIN

    SELECT INTO type_id id FROM auth_key_type WHERE auth_key_type.name = type;
    IF NOT FOUND THEN
        RAISE EXCEPTION 'Unknown key type "%"', type;
    END IF;

    DELETE FROM auth_key WHERE auth_key.type = type_id AND name = key_name;

    IF NOT FOUND THEN
        RAISE EXCEPTION 'No such % key "%"', type, key_name;
    END IF;

END;
$$ LANGUAGE plpgsql;
--
-- Support for requester proxying
--


DO $$
BEGIN
    IF NOT EXISTS (SELECT * FROM auth_key_type WHERE NAME = 'requester')
    THEN
        INSERT INTO auth_key_type (name) VALUES ('requester');
    END IF;
END;
$$ LANGUAGE plpgsql;
--
-- Priority
--


DO $$ BEGIN PERFORM drop_function_all('priority_default'); END $$;

CREATE OR REPLACE FUNCTION 
priority_default()
RETURNS INTEGER
AS $$
BEGIN
    RETURN 0;
END;
$$ LANGUAGE plpgsql
IMMUTABLE;


DO $$ BEGIN PERFORM drop_function_all('priority_min'); END $$;

CREATE OR REPLACE FUNCTION 
priority_min()
RETURNS INTEGER
AS $$
BEGIN
    RETURN -2147483648;
END;
$$ LANGUAGE plpgsql
IMMUTABLE;


DO $$ BEGIN PERFORM drop_function_all('priority_max'); END $$;

CREATE OR REPLACE FUNCTION 
priority_max()
RETURNS INTEGER
AS $$
BEGIN
    RETURN 2147483647;
END;
$$ LANGUAGE plpgsql
IMMUTABLE;

--
-- Table with one row that holds control values
--


DO $$
DECLARE
    t_name TEXT;            -- Name of the table being worked on
    t_version INTEGER;      -- Current version of the table
    t_version_old INTEGER;  -- Version of the table at the start
BEGIN

    --
    -- Preparation
    --

    t_name := 'control';

    t_version := table_version_find(t_name);
    t_version_old := t_version;


    --
    -- Upgrade Blocks
    --

    -- Version 0 (nonexistant) to version 1
    IF t_version = 0
    THEN

        CREATE TABLE control (

            -- End of run pause time
            pause_runs_until	TIMESTAMP WITH TIME ZONE

        );

        -- This table gets exactly one row that can only ever be updated.
        INSERT INTO control DEFAULT VALUES;

	t_version := t_version + 1;

    END IF;

    -- Version 1 to version 2
    -- IF t_version = 1
    -- THEN
    --     ,,,
    --     t_version := t_version + 1;
    -- END IF;


    --
    -- Cleanup
    --

    PERFORM table_version_set(t_name, t_version, t_version_old);

END;
$$ LANGUAGE plpgsql;



DROP TRIGGER IF EXISTS control_update ON control CASCADE;

DO $$ BEGIN PERFORM drop_function_all('control_update'); END $$;

CREATE OR REPLACE FUNCTION control_update()
RETURNS TRIGGER
AS $$
BEGIN
    NOTIFY control_changed;
    RETURN NEW;
END;
$$ LANGUAGE plpgsql;

CREATE TRIGGER control_update
AFTER UPDATE
ON control
FOR EACH ROW
    EXECUTE PROCEDURE control_update();



DROP TRIGGER IF EXISTS control_alter ON control CASCADE;
DROP TRIGGER IF EXISTS control_truncate ON control CASCADE;

DO $$ BEGIN PERFORM drop_function_all('control_noalter'); END $$;

CREATE OR REPLACE FUNCTION control_noalter()
RETURNS TRIGGER
AS $$
BEGIN
	RAISE EXCEPTION 'This table can only be updated.';
END;
$$ LANGUAGE plpgsql;

CREATE TRIGGER control_alter
BEFORE INSERT OR DELETE
ON control
FOR EACH ROW
    EXECUTE PROCEDURE control_noalter();

CREATE TRIGGER control_truncate
BEFORE TRUNCATE
ON control
EXECUTE PROCEDURE control_noalter();


--
-- API for Individual Columns
--

--
-- pause_runs_until
--

DO $$ BEGIN PERFORM drop_function_all('control_pause'); END $$;

CREATE OR REPLACE FUNCTION control_pause(
    duration INTERVAL DEFAULT NULL
)
RETURNS VOID
AS $$
DECLARE
    end_time TIMESTAMP WITH TIME ZONE;
BEGIN
    IF duration IS NOT NULL
    THEN
        end_time := now() + duration;
    ELSE
        end_time := tstz_infinity();
    END IF;

    UPDATE control SET pause_runs_until = end_time;
END;
$$ LANGUAGE plpgsql;


DO $$ BEGIN PERFORM drop_function_all('control_is_paused'); END $$;

CREATE OR REPLACE FUNCTION control_is_paused()
RETURNS BOOLEAN
AS $$
DECLARE
    is_paused BOOLEAN;
BEGIN
    SELECT INTO is_paused EXISTS (
        SELECT * FROM control WHERE pause_runs_until > now()
    );

    RETURN is_paused;
END;
$$ LANGUAGE plpgsql;


DO $$ BEGIN PERFORM drop_function_all('control_resume'); END $$;

CREATE OR REPLACE FUNCTION control_resume()
RETURNS VOID
AS $$
BEGIN
    UPDATE control SET pause_runs_until = NULL;
END;
$$ LANGUAGE plpgsql;
--
-- Queue of Pending HTTP Operations
--

-- This table holds a list of HTTP operations to be carried out until
-- successful.  One attempt to process each row after INSERT so the
-- query is handled immediately.  Failing queries will remain in the
-- table and retried at their specified interval until they expire.

DO $$
DECLARE
    t_name TEXT;            -- Name of the table being worked on
    t_version INTEGER;      -- Current version of the table
    t_version_old INTEGER;  -- Version of the table at the start
BEGIN

    --
    -- Preparation
    --

    t_name := 'http_queue';

    t_version := table_version_find(t_name);
    t_version_old := t_version;


    --
    -- Upgrade Blocks
    --

    -- Version 0 (nonexistant) to version 1
    IF t_version = 0
    THEN

        CREATE TABLE http_queue (

        	-- Row identifier
        	id		BIGSERIAL
        			PRIMARY KEY,

		operation	TEXT
        			CHECK (operation IN ('GET', 'PUT',
				                     'POST', 'DELETE')),

        	-- URI to fetch
        	uri		TEXT
        			NOT NULL,

        	-- Payload for GET and PUT
        	payload		TEXT
        			CHECK (NULL OR operation IN ('GET', 'PUT')),

        	-- How often to try after the first attempt
        	try_interval	INTERVAL
        			DEFAULT 'PT5M',

        	-- How long to keep trying.  This will be defaulted on INSERT.
        	try_for	     	INTERVAL,

        	-- When added
        	added		TIMESTAMP WITH TIME ZONE
        			DEFAULT now(),

        	-- When to stop trying
        	expires		TIMESTAMP WITH TIME ZONE,

        	-- Last time an attempt was made
        	last_attempt	TIMESTAMP WITH TIME ZONE,

        	-- Last status returned
        	last_status	INTEGER,

        	-- Last result returned
        	last_returned	TEXT,

        	-- Last time an attempt was made
        	next_attempt	TIMESTAMP WITH TIME ZONE
        			DEFAULT now(),

        	-- Attempts so far (really just for information)
        	attempts	INTEGER
        			DEFAULT 0

        );

	t_version := t_version + 1;

    END IF;

    -- Version 1 to version 2
    -- Adds 'timeout' column
    IF t_version = 1
    THEN
        ALTER TABLE http_queue ADD COLUMN
        timeout INTERVAL DEFAULT 'PT5S';

        UPDATE http_queue SET timeout = DEFAULT;

        t_version := t_version + 1;
    END IF;


    -- Version 2 to version 3
    -- Adds 'bind' column
    IF t_version = 2
    THEN
        ALTER TABLE http_queue ADD COLUMN
        bind TEXT NULL;

        t_version := t_version + 1;
    END IF;


    --
    -- Cleanup
    --

    PERFORM table_version_set(t_name, t_version, t_version_old);

END;
$$ LANGUAGE plpgsql;


-- Process one item in the table by its row ID

DO $$ BEGIN PERFORM drop_function_all('http_queue_process'); END $$;

CREATE OR REPLACE FUNCTION http_queue_process(
    row_id BIGINT
)
RETURNS VOID
AS $$
DECLARE
    entry RECORD;
    status http_result;
    status_family INTEGER;
    timeout_seconds FLOAT;
BEGIN

    SELECT INTO entry * from http_queue WHERE id = row_id;
    IF NOT FOUND
    THEN
        RETURN;
    END IF;

    timeout_seconds := extract('epoch' FROM entry.timeout);

    IF entry.operation = 'DELETE'
    THEN
 	status := http_delete(entry.uri, timeout_seconds, entry.bind);
    ELSEIF entry.operation = 'GET'
    THEN
        status := http_get(entry.uri, timeout_seconds, entry.bind);
    ELSEIF entry.operation = 'POST'
    THEN
 	status := http_post(entry.uri, entry.payload, timeout_seconds, entry.bind);
    ELSEIF entry.operation = 'PUT'
    THEN
 	status := http_put(entry.uri, entry.payload, timeout_seconds, entry.bind);
    ELSE
        RAISE EXCEPTION 'Unsupported operation %', entry.operation;
    END IF;

    status_family := (status.status/100)::INTEGER;

    IF status_family IN (1, 2, 3) -- Successful results
        OR now() + entry.try_interval > entry.expires
    THEN
        -- No need to keep succeeded or expired records around
	DELETE FROM http_queue WHERE id = entry.id;
    ELSE
        -- Set up the next attempt
        UPDATE http_queue
        SET
            attempts = attempts + 1,
 	    last_status = status.status,
 	    last_returned = status.returned,
            last_attempt = now(),
            next_attempt = now() + entry.try_interval
        WHERE id = row_id;
    END IF;
 
END;
$$ LANGUAGE plpgsql;

DO $$ BEGIN PERFORM drop_function_all('http_queue_process_all'); END $$;

CREATE OR REPLACE FUNCTION http_queue_process_all()
RETURNS VOID
AS $$
BEGIN
    PERFORM http_queue_process(id) FROM http_queue
    WHERE next_attempt < now() OR attempts = 0;
END;
$$ LANGUAGE plpgsql;



DROP TRIGGER IF EXISTS http_queue_insert ON http_queue CASCADE;

DO $$ BEGIN PERFORM drop_function_all('http_queue_insert'); END $$;

CREATE OR REPLACE FUNCTION http_queue_insert()
RETURNS TRIGGER
AS $$
BEGIN
    IF NEW.try_for IS NULL
    THEN
        SELECT INTO NEW.try_for keep_runs_tasks FROM configurables;
    END IF;
    NEW.expires = NEW.added + NEW.try_for;
    RETURN NEW;
END;
$$ LANGUAGE plpgsql;

CREATE TRIGGER http_queue_insert
BEFORE INSERT ON http_queue
FOR EACH ROW EXECUTE PROCEDURE http_queue_insert();




DROP TRIGGER IF EXISTS http_queue_insert_after ON http_queue CASCADE;

DO $$ BEGIN PERFORM drop_function_all('http_queue_insert_after'); END $$;

CREATE OR REPLACE FUNCTION http_queue_insert_after()
RETURNS TRIGGER
AS $$
BEGIN
    NOTIFY http_queue_new;
    RETURN NEW;
END;
$$ LANGUAGE plpgsql;

CREATE TRIGGER http_queue_insert_after
AFTER INSERT ON http_queue
FOR EACH ROW EXECUTE PROCEDURE http_queue_insert_after();
--
-- Scheduling Class
--

DO $$
DECLARE
    t_name TEXT;            -- Name of the table being worked on
    t_version INTEGER;      -- Current version of the table
    t_version_old INTEGER;  -- Version of the table at the start
BEGIN

    --
    -- Preparation
    --

    t_name := 'scheduling_class';

    t_version := table_version_find(t_name);
    t_version_old := t_version;


    --
    -- Upgrade Blocks
    --

    -- Version 0 (nonexistant) to version 1
    IF t_version = 0
    THEN

        CREATE TABLE scheduling_class (

        	-- Row identifier
        	id		INTEGER
        			PRIMARY KEY,

        	-- Display name
        	display		TEXT
        			UNIQUE NOT NULL,

        	-- Enumeration for use by programs
        	enum		TEXT
        			UNIQUE NOT NULL,

        	-- Whether or not tests in the class should be scheduled at
        	-- any time, regardless of exclusivity.  This is intended for
        	-- long-running jobs like powstream.
        	anytime         BOOLEAN,

        	-- Whether or not tests in the class get exclusive schedule
        	-- time
        	exclusive       BOOLEAN,

        	-- Whether or not tests in the class generate multiple
        	-- results.  For now, this is actually identical to the value
        	-- of 'anytime'.
        	multi_result    BOOLEAN
        );

	t_version := t_version + 1;

    END IF;

    -- Version 1 to version 2
    --IF t_version = 1
    --THEN
    --    ALTER TABLE ...
    --    t_version := t_version + 1;
    --END IF;


    --
    -- Cleanup
    --

    PERFORM table_version_set(t_name, t_version, t_version_old);

END;
$$ LANGUAGE plpgsql;




--
-- Functions that encapsulate the numeric values for each state
--

DO $$ BEGIN PERFORM drop_function_all('scheduling_class_background_multi'); END $$;

CREATE OR REPLACE FUNCTION scheduling_class_background_multi()
RETURNS INTEGER
AS $$
BEGIN
	RETURN 1;
END;
$$ LANGUAGE plpgsql;


DO $$ BEGIN PERFORM drop_function_all('scheduling_class_exclusive'); END $$;

CREATE OR REPLACE FUNCTION scheduling_class_exclusive()
RETURNS INTEGER
AS $$
BEGIN
	RETURN 2;
END;
$$ LANGUAGE plpgsql;


DO $$ BEGIN PERFORM drop_function_all('scheduling_class_normal'); END $$;

CREATE OR REPLACE FUNCTION scheduling_class_normal()
RETURNS INTEGER
AS $$
BEGIN
	RETURN 3;
END;
$$ LANGUAGE plpgsql;


DO $$ BEGIN PERFORM drop_function_all('scheduling_class_background'); END $$;

CREATE OR REPLACE FUNCTION scheduling_class_background()
RETURNS INTEGER
AS $$
BEGIN
	RETURN 4;
END;
$$ LANGUAGE plpgsql;



DROP TRIGGER IF EXISTS scheduling_class_alter ON scheduling_class CASCADE;

DO $$ BEGIN PERFORM drop_function_all('scheduling_class_alter'); END $$;

CREATE OR REPLACE FUNCTION scheduling_class_alter()
RETURNS TRIGGER
AS $$
BEGIN
	RAISE EXCEPTION 'This table may not be altered';
	RETURN NEW;
END;
$$ LANGUAGE plpgsql;

CREATE TRIGGER scheduling_class_alter
BEFORE INSERT OR UPDATE OR DELETE
ON scheduling_class
       FOR EACH ROW EXECUTE PROCEDURE scheduling_class_alter();


-- Do this after trigger creation and with a DISABLE/ENABLE in case
-- the table was previously populated.

ALTER TABLE scheduling_class DISABLE TRIGGER scheduling_class_alter;
INSERT INTO scheduling_class (id, display, enum, anytime, exclusive, multi_result)
VALUES
    (scheduling_class_background_multi(), 'Background Multi-Result',  'background-multi', TRUE,  FALSE, TRUE ),
    (scheduling_class_background(),       'Background Single-Result', 'background',       TRUE,  FALSE, FALSE),
    (scheduling_class_exclusive(),        'Exclusive',                'exclusive',        FALSE, TRUE,  FALSE),
    (scheduling_class_normal(),           'Normal',                   'normal',           FALSE, FALSE, FALSE)
ON CONFLICT (id) DO UPDATE
SET
    display = EXCLUDED.display,
    enum = EXCLUDED.enum,
    anytime = EXCLUDED.anytime,
    exclusive = EXCLUDED.exclusive,
    multi_result = EXCLUDED.multi_result;
ALTER TABLE scheduling_class ENABLE TRIGGER scheduling_class_alter;


-- Find the ID for a class by its enumeration

DO $$ BEGIN PERFORM drop_function_all('scheduling_class_find'); END $$;

CREATE OR REPLACE FUNCTION scheduling_class_find(
    candidate TEXT
)
RETURNS INTEGER
AS $$
BEGIN
    RETURN (SELECT id FROM scheduling_class WHERE enum = candidate);
END;
$$ LANGUAGE plpgsql;
--
-- Table of Test Types
--

DO $$
DECLARE
    t_name TEXT;            -- Name of the table being worked on
    t_version INTEGER;      -- Current version of the table
    t_version_old INTEGER;  -- Version of the table at the start
BEGIN

    --
    -- Preparation
    --

    t_name := 'test';

    t_version := table_version_find(t_name);
    t_version_old := t_version;


    --
    -- Upgrade Blocks
    --

    -- Version 0 (nonexistant) to version 1
    IF t_version = 0
    THEN

        CREATE TABLE test (

        	-- Row identifier
        	id		BIGSERIAL
        			PRIMARY KEY,

        	-- Original JSON
        	json		JSONB
        			NOT NULL,

        	-- Test Name
        	name		TEXT
        			UNIQUE NOT NULL,

        	-- Verbose description
        	description	TEXT,

        	-- Version
        	version		NUMERIC
        			NOT NULL,

        	-- When this record was last updated
        	updated		TIMESTAMP WITH TIME ZONE,

        	-- Whether or not the test is currently available
        	available	BOOLEAN
        			DEFAULT TRUE,

                -- Scheduling class for this type of test
        	scheduling_class INTEGER
                                NOT NULL
        			REFERENCES scheduling_class(id)
        			-- Deletes should never happen but for consistency...
        			ON DELETE CASCADE
        );

	t_version := t_version + 1;

    END IF;

    -- Version 1 to version 2
    -- Remove unused and trouble-causing version column
    IF t_version = 1
    THEN
        ALTER TABLE test DROP COLUMN version;

        t_version := t_version + 1;
    END IF;


    --
    -- Cleanup
    --

    PERFORM table_version_set(t_name, t_version, t_version_old);

END;
$$ LANGUAGE plpgsql;


DROP TRIGGER IF EXISTS test_alter ON test CASCADE;

DO $$ BEGIN PERFORM drop_function_all('test_alter'); END $$;

CREATE OR REPLACE FUNCTION test_alter()
RETURNS TRIGGER
AS $$
DECLARE
    json_result TEXT;
BEGIN

    json_result := json_validate(NEW.json, '#/pScheduler/PluginEnumeration/Test');
    IF json_result IS NOT NULL
    THEN
        RAISE EXCEPTION 'Invalid enumeration: %', json_result;
    END IF;

    NEW.name := NEW.json ->> 'name';
    NEW.description := NEW.json ->> 'description';
    NEW.scheduling_class := scheduling_class_find(NEW.json ->> 'scheduling-class');

    RETURN NEW;
END;
$$ LANGUAGE plpgsql;

CREATE TRIGGER test_alter
BEFORE INSERT OR UPDATE
ON test
FOR EACH ROW
    EXECUTE PROCEDURE test_alter();



-- Function to run at startup.

DO $$ BEGIN PERFORM drop_function_all('test_boot'); END $$;

CREATE OR REPLACE FUNCTION test_boot()
RETURNS VOID
AS $$
DECLARE
    run_result external_program_result;
    test_list JSONB;
    test_name TEXT;
    test_enumeration JSONB;
    json_result TEXT;
    sschema NUMERIC;  -- Name dodges a reserved word
BEGIN
    run_result := pscheduler_command(ARRAY['internal', 'list', 'test']);
    IF run_result.status <> 0 THEN
       RAISE EXCEPTION 'Unable to list installed tests: %', run_result.stderr;
    END IF;

    test_list := run_result.stdout::JSONB;

    FOR test_name IN (select * from jsonb_array_elements_text(test_list))
    LOOP
	run_result := pscheduler_plugin_invoke('test', test_name, 'enumerate');
        IF run_result.status <> 0 THEN
            RAISE WARNING 'Test "%" failed to enumerate: %',
	        test_name, run_result.stderr;
            CONTINUE;
        END IF;

	test_enumeration := run_result.stdout::JSONB;

        sschema := text_to_numeric(test_enumeration ->> 'schema');
        IF sschema IS NOT NULL AND sschema > 1 THEN
            RAISE WARNING 'Test "%": schema % is not supported',
                test_name, sschema;
            CONTINUE;
        END IF;

        json_result := json_validate(test_enumeration,
	    '#/pScheduler/PluginEnumeration/Test');
        IF json_result IS NOT NULL
        THEN
            RAISE WARNING 'Invalid enumeration for test "%": %', test_name, json_result;
	    CONTINUE;
        END IF;

	INSERT INTO test (json, updated, available)
	VALUES (test_enumeration, now(), TRUE)
	ON CONFLICT (name) DO UPDATE
        SET json = test_enumeration, updated = now(), available = TRUE;

    END LOOP;

    -- Disable, but don't remove, tests that aren't installed.
    UPDATE test SET available = FALSE WHERE updated < now();
    -- TODO: Should also can anything on the schedule.  (Do that elsewhere.)
END;
$$ LANGUAGE plpgsql;
--
-- Tables of tools and the tests they run.
--

DO $$
DECLARE
    t_name TEXT;            -- Name of the table being worked on
    t_version INTEGER;      -- Current version of the table
    t_version_old INTEGER;  -- Version of the table at the start
BEGIN

    --
    -- Preparation
    --

    t_name := 'tool';

    t_version := table_version_find(t_name);
    t_version_old := t_version;


    --
    -- Upgrade Blocks
    --

    -- Version 0 (nonexistant) to version 1
    IF t_version = 0
    THEN

        CREATE TABLE tool (

        	-- Row identifier
        	id		BIGSERIAL
        			PRIMARY KEY,

        	-- Original JSON
        	json		JSONB
        			NOT NULL,

        	-- Tool Name
        	name		TEXT
        			UNIQUE NOT NULL,

        	-- Verbose description
        	description	TEXT,

        	-- Version
        	version		NUMERIC
        			NOT NULL,

        	-- Preference value
        	preference      INTEGER
        			DEFAULT 0,

        	-- When this record was last updated
        	updated		TIMESTAMP WITH TIME ZONE,

        	-- Whether or not the tool is currently available
        	available	BOOLEAN
        			DEFAULT TRUE
        );

        CREATE INDEX tool_name ON tool(name);

	t_version := t_version + 1;

    END IF;

    -- Version 1 to version 2
    -- Remove unused and trouble-causing version column
    IF t_version = 1
    THEN
        ALTER TABLE tool DROP COLUMN version;

        t_version := t_version + 1;
    END IF;


    --
    -- Cleanup
    --

    PERFORM table_version_set(t_name, t_version, t_version_old);

END;
$$ LANGUAGE plpgsql;


--
-- Breaker table that maps tools to the tests they can run
--


DO $$
DECLARE
    t_name TEXT;            -- Name of the table being worked on
    t_version INTEGER;      -- Current version of the table
    t_version_old INTEGER;  -- Version of the table at the start
BEGIN

    --
    -- Preparation
    --

    t_name := 'tool_test';

    t_version := table_version_find(t_name);
    t_version_old := t_version;


    --
    -- Upgrade Blocks
    --

    -- Version 0 (nonexistant) to version 1
    IF t_version = 0
    THEN

        CREATE TABLE tool_test (

        	-- Tool which says it can handle a test
        	tool		INTEGER
        			REFERENCES tool(id)
        			ON DELETE CASCADE,

        	-- The test the tool says it can handle
        	test		INTEGER
        			REFERENCES test(id)
        			ON DELETE CASCADE
        );

	t_version := t_version + 1;

    END IF;

    -- Version 1 to version 2
    --IF t_version = 1
    --THEN
    --    ALTER TABLE ...
    --    t_version := t_version + 1;
    --END IF;


    --
    -- Cleanup
    --

    PERFORM table_version_set(t_name, t_version, t_version_old);

END;
$$ LANGUAGE plpgsql;



DROP TRIGGER IF EXISTS tool_alter ON tool CASCADE;

CREATE OR REPLACE FUNCTION tool_alter()
RETURNS TRIGGER
AS $$
DECLARE
    json_result TEXT;
BEGIN

    json_result := json_validate(NEW.json, '#/pScheduler/PluginEnumeration/Tool');
    IF json_result IS NOT NULL
    THEN
        RAISE EXCEPTION 'Invalid enumeration: %', json_result;
    END IF;

    NEW.name := NEW.json ->> 'name';
    NEW.description := NEW.json ->> 'description';
    NEW.preference := text_to_numeric(NEW.json ->> 'preference');
    RETURN NEW;
END;
$$ LANGUAGE plpgsql;

CREATE TRIGGER tool_alter
BEFORE INSERT OR UPDATE
ON tool
FOR EACH ROW
    EXECUTE PROCEDURE tool_alter();



DROP TRIGGER IF EXISTS tool_alter_post ON tool CASCADE;

CREATE OR REPLACE FUNCTION tool_alter_post()
RETURNS TRIGGER
AS $$
DECLARE
    test_name TEXT;
    test_id INTEGER;
BEGIN

    -- Update the breaker table between this and test.

    DELETE FROM tool_test WHERE tool = NEW.id;

    FOR test_name IN
        (SELECT * FROM jsonb_array_elements_text(NEW.json -> 'tests'))
    LOOP
        -- Only insert records for tests that are installed on the system.
        SELECT id INTO test_id FROM test WHERE name = test_name;
	IF FOUND THEN
	    INSERT INTO tool_test (tool, test)
	        VALUES (NEW.id, test_id);
        END IF;
    END LOOP;

    RETURN NEW;
END;
$$ LANGUAGE plpgsql;

CREATE TRIGGER tool_alter_post
AFTER INSERT OR UPDATE
ON tool
FOR EACH ROW
    EXECUTE PROCEDURE tool_alter_post();



DROP TRIGGER IF EXISTS tool_delete ON tool CASCADE;

CREATE OR REPLACE FUNCTION tool_delete()
RETURNS TRIGGER
AS $$
BEGIN
    DELETE FROM tool_test where tool = OLD.id;
    RETURN OLD;
END;
$$ LANGUAGE plpgsql;

CREATE TRIGGER tool_delete
BEFORE DELETE
ON tool
FOR EACH ROW
    EXECUTE PROCEDURE tool_delete();



-- Function to run at startup.

DO $$ BEGIN PERFORM drop_function_all('tool_boot'); END $$;

CREATE OR REPLACE FUNCTION tool_boot()
RETURNS VOID
AS $$
DECLARE
    run_result external_program_result;
    tool_list JSONB;
    tool_name TEXT;
    tool_enumeration JSONB;
    json_result TEXT;
    sschema NUMERIC;  -- Name dodges a reserved word
BEGIN
    run_result := pscheduler_command(ARRAY['internal', 'list', 'tool']);
    IF run_result.status <> 0 THEN
       RAISE EXCEPTION 'Unable to list installed tools: %', run_result.stderr;
    END IF;

    tool_list := run_result.stdout::JSONB;

    FOR tool_name IN (select * from jsonb_array_elements_text(tool_list))
    LOOP

	run_result := pscheduler_plugin_invoke('tool', tool_name, 'enumerate');
        IF run_result.status <> 0 THEN
            RAISE WARNING 'Tool "%" failed to enumerate: %',
	        tool_name, run_result.stderr;
            CONTINUE;
        END IF;

	tool_enumeration := run_result.stdout::JSONB;

        sschema := text_to_numeric(tool_enumeration ->> 'schema');
        IF sschema IS NOT NULL AND sschema > 1 THEN
            RAISE WARNING 'Tool "%": schema % is not supported',
                tool_name, sschema;
            CONTINUE;
        END IF;

        json_result := json_validate(tool_enumeration,
	    '#/pScheduler/PluginEnumeration/Tool');
        IF json_result IS NOT NULL
        THEN
            RAISE WARNING 'Invalid enumeration for tool "%": %', tool_name, json_result;
	    CONTINUE;
        END IF;

	INSERT INTO tool (json, updated, available)
	VALUES (tool_enumeration, now(), TRUE)
	ON CONFLICT (name) DO UPDATE
        SET json = tool_enumeration, updated = now(), available = TRUE;

    END LOOP;

    -- TODO: Disable, but don't remove, tools that aren't installed.
    UPDATE tool SET available = FALSE WHERE updated < now();
    -- TODO: Should also can anything on the schedule.  (Do that elsewhere.)
END;
$$ LANGUAGE plpgsql;



-- Determine whether or not a tool is willing to run a specific test.

DO $$ BEGIN PERFORM drop_function_all('tool_can_run_test'); END $$;

CREATE OR REPLACE FUNCTION tool_can_run_test(
    tool_id BIGINT,
    test JSONB
)
RETURNS JSONB
AS $$
DECLARE
    tool_name TEXT;
    run_result external_program_result;
    result_json JSONB;
BEGIN

    SELECT INTO tool_name name FROM tool WHERE id = tool_id;
    IF NOT FOUND THEN
        RAISE EXCEPTION 'Tool ID % is invalid', tool_id;
    END IF;

    run_result := pscheduler_plugin_invoke('tool', tool_name, 'can-run', test::TEXT, 5);


    -- Any result other than 1 indicates a problem that shouldn't be
    -- allowed to gum up the works.  Log it and assume the tool said
    -- no dice.
    IF run_result.status <> 0 THEN
        RAISE WARNING 'Tool "%" failed can-run: %', tool_name, run_result.stderr;
        RETURN '{ "can-run": false, "reasons": [ "Failed can-run; see system logs." ] }'::JSONB;
    END IF;

    result_json = text_to_jsonb(run_result.stdout);
    IF result_json IS NULL THEN
        RAISE WARNING 'Tool "%" returned invalid JSON "%"', tool_name, run_result.stdout;
	RETURN '{ "can-run": false, "reasons": [ "Failed can-run; see system logs." ] }'::JSONB;
    END IF;

    RETURN result_json;

END;
$$ LANGUAGE plpgsql;



--
-- API
--

-- Get a JSON array of the enumerations of all tools that can run a
-- test, returned in order of highest to lowest preference.


DO $$ BEGIN PERFORM drop_function_all('api_tools_for_test'); END $$;

CREATE OR REPLACE FUNCTION api_tools_for_test(
    test_json JSONB
)
RETURNS TABLE (
    can_run JSONB,
    tool JSONB
)
AS $$
DECLARE
    test_type TEXT;
BEGIN

    test_type := test_json ->> 'type';
    IF test_type IS NULL THEN
        RAISE EXCEPTION 'No test type found in JSON';
    END IF;

    RETURN QUERY
    SELECT
        tool_can_run_test( tool.id, test_json )::JSONB as can_run,
        tool.json::JSONB AS tool
    FROM
        test
        JOIN tool_test ON tool_test.test = test.id
        JOIN tool ON tool.id = tool_test.tool
    WHERE
        test.name = test_type
        AND test.available
        AND tool.available
    ORDER BY
        tool.preference DESC,
        tool.name ASC
    ;

END;
$$ LANGUAGE plpgsql;
--
-- Table of archives to be used with every run
--

DO $$
DECLARE
    t_name TEXT;            -- Name of the table being worked on
    t_version INTEGER;      -- Current version of the table
    t_version_old INTEGER;  -- Version of the table at the start
BEGIN

    --
    -- Preparation
    --

    t_name := 'archive_default';

    t_version := table_version_find(t_name);
    t_version_old := t_version;


    --
    -- Upgrade Blocks
    --

    -- Version 0 (nonexistant) to version 1
    IF t_version = 0
    THEN

        CREATE TABLE archive_default (

            -- Archive specification
            archive             JSONB,

            -- When the record was inserted (for debug only)
            inserted            TIMESTAMP WITH TIME ZONE
                                DEFAULT now()
        );

	t_version := t_version + 1;

    END IF;

    -- Version 1 to version 2
    --IF t_version = 1
    --THEN
    --    ALTER TABLE ...
    --    t_version := t_version + 1;
    --END IF;


    --
    -- Cleanup
    --

    PERFORM table_version_set(t_name, t_version, t_version_old);

END;
$$ LANGUAGE plpgsql;


DROP TRIGGER IF EXISTS archive_default_insert ON archive_default CASCADE;

DO $$ BEGIN PERFORM drop_function_all('archive_default_insert'); END $$;

CREATE OR REPLACE FUNCTION archive_default_insert()
RETURNS TRIGGER
AS $$
BEGIN
    PERFORM archiver_validate(NEW.archive);
    RETURN NEW;
END;
$$ LANGUAGE plpgsql;

CREATE TRIGGER archive_default_insert
AFTER INSERT
ON archive_default
FOR EACH ROW
    EXECUTE PROCEDURE archive_default_insert();



-- Rows in this can't be updated, only inserted and deleted.  (And
-- usually the whole table at once.)

DROP TRIGGER IF EXISTS archive_default_update ON archive_default CASCADE;

DO $$ BEGIN PERFORM drop_function_all('archive_default_update'); END $$;

CREATE OR REPLACE FUNCTION archive_default_update()
RETURNS TRIGGER
AS $$
BEGIN
    RAISE EXCEPTION 'Rows in this table cannot be updated.';
END;
$$ LANGUAGE plpgsql;

CREATE TRIGGER archive_default_update
AFTER UPDATE
ON archive_default
FOR EACH ROW
    EXECUTE PROCEDURE archive_default_update();
--
-- Table of archivers
--


DO $$
DECLARE
    t_name TEXT;            -- Name of the table being worked on
    t_version INTEGER;      -- Current version of the table
    t_version_old INTEGER;  -- Version of the table at the start
BEGIN

    --
    -- Preparation
    --

    t_name := 'archiver';

    t_version := table_version_find(t_name);
    t_version_old := t_version;


    --
    -- Upgrade Blocks
    --

    -- Version 0 (nonexistant) to version 1
    IF t_version = 0
    THEN

        CREATE TABLE archiver (

        	-- Row identifier
        	id		BIGSERIAL
        			PRIMARY KEY,

        	-- Original JSON
        	json		JSONB
        			NOT NULL,

        	-- Archiver Name
        	name		TEXT
        			UNIQUE NOT NULL,

        	-- Verbose description
        	description	TEXT,

        	-- Version
        	version		NUMERIC
        			NOT NULL,

        	-- When this record was last updated
        	updated		TIMESTAMP WITH TIME ZONE,

        	-- Whether or not the archiver is currently available
        	available	BOOLEAN
        			DEFAULT TRUE
        );


        CREATE INDEX archiver_name ON archiver(name);

	t_version := t_version + 1;

    END IF;

    -- Version 1 to version 2
    -- Remove unused and trouble-causing version column
    IF t_version = 1
    THEN
        ALTER TABLE archiver DROP COLUMN version;

        t_version := t_version + 1;
    END IF;


    --
    -- Cleanup
    --

    PERFORM table_version_set(t_name, t_version, t_version_old);

END;
$$ LANGUAGE plpgsql;


--
-- Breaker table that maps archivers to the tests they can run
--


DO $$
DECLARE
    t_name TEXT;            -- Name of the table being worked on
    t_version INTEGER;      -- Current version of the table
    t_version_old INTEGER;  -- Version of the table at the start
BEGIN

    --
    -- Preparation
    --

    t_name := 'archiver_test';

    t_version := table_version_find(t_name);
    t_version_old := t_version;


    --
    -- Upgrade Blocks
    --

    -- Version 0 (nonexistant) to version 1
    IF t_version = 0
    THEN

        CREATE TABLE archiver_test (

        	-- Archiver which says it can handle a test
        	archiver	BIGINT
        			REFERENCES archiver(id)
        			ON DELETE CASCADE,

        	-- The test the archiver says it can handle
        	test		BIGINT
        			REFERENCES test(id)
        			ON DELETE CASCADE
        );

	t_version := t_version + 1;

    END IF;

    -- Version 1 to version 2
    --IF t_version = 1
    --THEN
    --    ALTER TABLE ...
    --    t_version := t_version + 1;
    --END IF;


    --
    -- Cleanup
    --

    PERFORM table_version_set(t_name, t_version, t_version_old);

END;
$$ LANGUAGE plpgsql;



DROP TRIGGER IF EXISTS archiver_alter ON archiver CASCADE;

DO $$ BEGIN PERFORM drop_function_all('archiver_alter'); END $$;

CREATE OR REPLACE FUNCTION archiver_alter()
RETURNS TRIGGER
AS $$
DECLARE
    json_result TEXT;
BEGIN
    json_result := json_validate(NEW.json, '#/pScheduler/PluginEnumeration/Archiver');
    IF json_result IS NOT NULL
    THEN
        RAISE EXCEPTION 'Invalid enumeration: %', json_result;
    END IF;

    NEW.name := NEW.json ->> 'name';
    NEW.description := NEW.json ->> 'description';
    RETURN NEW;
END;
$$ LANGUAGE plpgsql;

CREATE TRIGGER archiver_alter
BEFORE INSERT OR UPDATE
ON archiver
FOR EACH ROW
    EXECUTE PROCEDURE archiver_alter();



DROP TRIGGER IF EXISTS archiver_alter_post ON archiver CASCADE;

DO $$ BEGIN PERFORM drop_function_all('archiver_alter_post'); END $$;

CREATE OR REPLACE FUNCTION archiver_alter_post()
RETURNS TRIGGER
AS $$
DECLARE
    test_name TEXT;
    test_id BIGINT;
BEGIN

    -- Update the breaker table between this and test.

    DELETE FROM archiver_test WHERE archiver = NEW.id;

    FOR test_name IN
        (SELECT * FROM jsonb_array_elements_text(NEW.json -> 'tests'))
    LOOP
        -- Only insert records for tests that are installed on the system.
        SELECT id INTO test_id FROM test WHERE name = test_name;
	IF FOUND THEN
	    INSERT INTO archiver_test (archiver, test)
	        VALUES (NEW.id, test_id);
        END IF;
    END LOOP;

    RETURN NEW;
END;
$$ LANGUAGE plpgsql;

CREATE TRIGGER archiver_alter_post
AFTER INSERT OR UPDATE
ON archiver
FOR EACH ROW
    EXECUTE PROCEDURE archiver_alter_post();


DROP TRIGGER IF EXISTS archiver_delete ON archiver CASCADE;

DO $$ BEGIN PERFORM drop_function_all('archiver_delete'); END $$;

CREATE OR REPLACE FUNCTION archiver_delete()
RETURNS TRIGGER
AS $$
BEGIN
    DELETE FROM archiver_test where archiver = OLD.id;
    RETURN OLD;
END;
$$ LANGUAGE plpgsql;

CREATE TRIGGER archiver_delete
BEFORE DELETE
ON archiver
FOR EACH ROW
    EXECUTE PROCEDURE archiver_delete();



-- Function to run at startup.

DO $$ BEGIN PERFORM drop_function_all('archiver_boot'); END $$;

CREATE OR REPLACE FUNCTION archiver_boot()
RETURNS VOID
AS $$
DECLARE
    run_result external_program_result;
    archiver_list JSONB;
    archiver_name TEXT;
    archiver_enumeration JSONB;
    json_result TEXT;
    sschema NUMERIC;  -- Name dodges a reserved word
BEGIN
    run_result := pscheduler_command(ARRAY['internal', 'list', 'archiver']);
    IF run_result.status <> 0 THEN
       RAISE EXCEPTION 'Unable to list installed archivers: %', run_result.stderr;
    END IF;

    archiver_list := run_result.stdout::JSONB;

    FOR archiver_name IN (select * from jsonb_array_elements_text(archiver_list))
    LOOP

	run_result := pscheduler_plugin_invoke('archiver', archiver_name, 'enumerate');
        IF run_result.status <> 0 THEN
            RAISE WARNING 'Archiver "%" failed to enumerate: %',
	        archiver_name, run_result.stderr;
            CONTINUE;
        END IF;

	archiver_enumeration := run_result.stdout::JSONB;

        sschema := text_to_numeric(archiver_enumeration ->> 'schema');
        IF sschema IS NOT NULL AND sschema > 1 THEN
            RAISE WARNING 'Archiver "%": schema % is not supported',
                archiver_name, sschema;
            CONTINUE;
        END IF;

        json_result := json_validate(archiver_enumeration,
	    '#/pScheduler/PluginEnumeration/Archiver');
        IF json_result IS NOT NULL
        THEN
            RAISE WARNING 'Invalid enumeration for archiver "%": %', archiver_name, json_result;
	    CONTINUE;
        END IF;

	INSERT INTO archiver (json, updated, available)
	VALUES (archiver_enumeration, now(), TRUE)
	ON CONFLICT (name) DO UPDATE
        SET json = archiver_enumeration, updated = now(), available = TRUE;

    END LOOP;

    -- TODO: Disable, but don't remove, archivers that aren't installed.
    UPDATE archiver SET available = FALSE WHERE updated < now();
    -- TODO: Should also can anything on the schedule that used this archiver.  (Do that elsewhere.)
END;
$$ LANGUAGE plpgsql;



-- Validate an archiver entry and raise an error if invalid.

DO $$ BEGIN PERFORM drop_function_all('archiver_validate'); END $$;

CREATE OR REPLACE FUNCTION archiver_validate(
    candidate JSONB
)
RETURNS VOID
AS $$
DECLARE
    archiver_name TEXT;
    candidate_data JSONB;
    run_result external_program_result;
    validate_result JSONB;
BEGIN
    IF NOT candidate ? 'archiver' THEN
        RAISE EXCEPTION 'No archiver name specified in archiver.';
    END IF;

    archiver_name := candidate ->> 'archiver';

    IF NOT EXISTS (SELECT * FROM archiver WHERE name = archiver_name) THEN
        RAISE EXCEPTION 'No archiver "%" is avaiable.', archiver_name;
    END IF;

    IF candidate ? 'data' THEN
        candidate_data := candidate -> 'data';
    ELSE
        candidate_data := 'null'::JSONB;
    END IF;

    run_result := pscheduler_plugin_invoke('archiver', archiver_name, 'data-is-valid',
        candidate_data::TEXT );
    IF run_result.status <> 0 THEN
        RAISE EXCEPTION 'Archiver "%" failed to validate: %', archiver_name, run_result.stderr;
    END IF;

    validate_result := run_result.stdout::JSONB;

    IF NOT (validate_result ->> 'valid')::BOOLEAN THEN
        IF validate_result ? 'error' THEN
            RAISE EXCEPTION 'Invalid data for archiver "%": %', archiver_name, validate_result ->> 'error';
        ELSE
            RAISE EXCEPTION 'Invalid data for archiver "%": No error provided by plugin.', archiver_name;
        END IF;
    END IF;

END;
$$ LANGUAGE plpgsql;



-- Quick summary

CREATE OR REPLACE VIEW archiver_summary
AS
    SELECT
        archiver.name AS archiver,
        test.name AS test
    FROM
        archiver_test
	JOIN archiver ON archiver.id = archiver
	JOIN test ON test.id = test
;
--
-- Table of contexts
--

DO $$
DECLARE
    t_name TEXT;            -- Name of the table being worked on
    t_version INTEGER;      -- Current version of the table
    t_version_old INTEGER;  -- Version of the table at the start
BEGIN

    --
    -- Preparation
    --

    t_name := 'context';

    t_version := table_version_find(t_name);
    t_version_old := t_version;


    --
    -- Upgrade Blocks
    --

    -- Version 0 (nonexistant) to version 1
    IF t_version = 0
    THEN

        CREATE TABLE context (

        	-- Row identifier
        	id		BIGSERIAL
        			PRIMARY KEY,

        	-- Original JSON
        	json		JSONB
        			NOT NULL,

        	-- Context Name
        	name		TEXT
        			UNIQUE NOT NULL,

        	-- Verbose description
        	description	TEXT,

        	-- When this record was last updated
        	updated		TIMESTAMP WITH TIME ZONE,

        	-- Whether or not the context is currently available
        	available	BOOLEAN
        			DEFAULT TRUE
        );


        CREATE INDEX context_name ON context(name);

	t_version := t_version + 1;

    END IF;

    --
    -- Cleanup
    --

    PERFORM table_version_set(t_name, t_version, t_version_old);

END;
$$ LANGUAGE plpgsql;



DROP TRIGGER IF EXISTS context_alter ON context CASCADE;

DO $$ BEGIN PERFORM drop_function_all('context_alter'); END $$;

CREATE OR REPLACE FUNCTION context_alter()
RETURNS TRIGGER
AS $$
DECLARE
    json_result TEXT;
BEGIN
    -- TODO: Need to add this to the dictionary
    json_result := json_validate(NEW.json, '#/pScheduler/PluginEnumeration/Context');
    IF json_result IS NOT NULL
    THEN
        RAISE EXCEPTION 'Invalid enumeration: %', json_result;
    END IF;

    NEW.name := NEW.json ->> 'name';
    NEW.description := NEW.json ->> 'description';
    RETURN NEW;
END;
$$ LANGUAGE plpgsql;

CREATE TRIGGER context_alter
BEFORE INSERT OR UPDATE
ON context
FOR EACH ROW
    EXECUTE PROCEDURE context_alter();



DROP TRIGGER IF EXISTS context_alter_post ON context CASCADE;


-- Function to run at startup.

DO $$ BEGIN PERFORM drop_function_all('context_boot'); END $$;

CREATE OR REPLACE FUNCTION context_boot()
RETURNS VOID
AS $$
DECLARE
    run_result external_program_result;
    context_list JSONB;
    context_name TEXT;
    context_enumeration JSONB;
    json_result TEXT;
    sschema NUMERIC;  -- Name dodges a reserved word
BEGIN
    run_result := pscheduler_command(ARRAY['internal', 'list', 'context']);
    IF run_result.status <> 0 THEN
       RAISE EXCEPTION 'Unable to list installed contexts: %', run_result.stderr;
    END IF;

    context_list := run_result.stdout::JSONB;

    FOR context_name IN (select * from jsonb_array_elements_text(context_list))
    LOOP

	run_result := pscheduler_plugin_invoke('context', context_name, 'enumerate');
        IF run_result.status <> 0 THEN
            RAISE WARNING 'Context "%" failed to enumerate: %',
	        context_name, run_result.stderr;
            CONTINUE;
        END IF;

	context_enumeration := run_result.stdout::JSONB;

        sschema := text_to_numeric(context_enumeration ->> 'schema');
        IF sschema IS NOT NULL AND sschema > 1 THEN
            RAISE WARNING 'Context "%": schema % is not supported',
                context_name, sschema;
            CONTINUE;
        END IF;

        json_result := json_validate(context_enumeration,
	    '#/pScheduler/PluginEnumeration/Context');
        IF json_result IS NOT NULL
        THEN
            RAISE WARNING 'Invalid enumeration for context "%": %', context_name, json_result;
	    CONTINUE;
        END IF;

	INSERT INTO context (json, updated, available)
        VALUES (context_enumeration, now(), true)
        ON CONFLICT (name) DO UPDATE
        SET json = context_enumeration, updated = now(), available = TRUE;

    END LOOP;

    -- TODO: Disable, but don't remove, contexts that aren't installed.
    UPDATE context SET available = FALSE WHERE updated < now();
    -- TODO: Should also can anything on the schedule that used this context.  (Do that elsewhere.)
END;
$$ LANGUAGE plpgsql;



-- Validate a context entry and raise an error if invalid.

DO $$ BEGIN PERFORM drop_function_all('context_validate'); END $$;

CREATE OR REPLACE FUNCTION context_validate(
    candidate JSONB
)
RETURNS VOID
AS $$
DECLARE
    json_result TEXT;
    context_name TEXT;
    run_result external_program_result;
    validate_result JSONB;
BEGIN

    json_result := json_validate(candidate, '#/pScheduler/ContextSpecificationSingle');
    IF json_result IS NOT NULL
    THEN
        RAISE EXCEPTION 'Invalid context specification: %', json_result;
    END IF;

    context_name := (candidate ->> 'context')::TEXT;

    IF NOT EXISTS (SELECT * FROM context WHERE name = context_name)
    THEN
        RAISE EXCEPTION 'Unknown context "%"', context_name;
    END IF;

    run_result := pscheduler_plugin_invoke('context', context_name, 'data-is-valid',
        (candidate ->> 'data')::TEXT );
    IF run_result.status <> 0 THEN
        RAISE EXCEPTION 'Context %/% failed to validate: %',
	    context_name, (candidate ->> 'data')::TEXT, run_result.stderr;
    END IF;

    validate_result := run_result.stdout::JSONB;

    IF NOT (validate_result ->> 'valid')::BOOLEAN THEN
        IF validate_result ? 'error' THEN
            RAISE EXCEPTION 'Invalid data for context "%": %',
	        context_name, validate_result ->> 'error';
        ELSE
            RAISE EXCEPTION 'Invalid data for context "%": No error provided by plugin.',
	        context_name;
        END IF;
    END IF;

END;
$$ LANGUAGE plpgsql;
--
-- Task Table
--

DO $$
DECLARE
    t_name TEXT;            -- Name of the table being worked on
    t_version INTEGER;      -- Current version of the table
    t_version_old INTEGER;  -- Version of the table at the start

    -- Used by version 1 to 2 upgrade
    v1_2_record RECORD;
    v1_2_run_result external_program_result;
BEGIN

    --
    -- Preparation
    --

    t_name := 'task';

    t_version := table_version_find(t_name);
    t_version_old := t_version;


    --
    -- Upgrade Blocks
    --

    -- Version 0 (nonexistant) to version 1
    IF t_version = 0
    THEN

        CREATE TABLE task (

        	-- Row identifier
        	id		BIGSERIAL
        			PRIMARY KEY,

        	-- External-use identifier
        	uuid		UUID
        			UNIQUE
                                NOT NULL,

        	-- JSON representation of the task.  This should be the only
        	-- column specified when inserting a row; all of the others
        	-- are derived and will be overwritten.
        	json		JSONB
        			NOT NULL,

        	--
        	-- TEST
        	--

        	test		BIGINT
        			REFERENCES test(id)
        			ON DELETE CASCADE,

        	--
        	-- SCHEDULING
        	--

        	-- When this task was first added to the schedule.  This is
        	-- used for breaking ties on equal-priority tasks.
        	added		TIMESTAMP WITH TIME ZONE,

        	-- When the first run of the test should start.  If NULL, the
        	-- scheduler should find the first available time for the
        	-- first iteration and use that.
        	start		TIMESTAMP WITH TIME ZONE
        			DEFAULT NULL,

        	-- Amount of slip permissible in start time
        	slip		INTERVAL,

        	-- How much of the slip is randomly applied to the start time.
		-- NOTE: This column was remove in version 9.
        	randslip	NUMERIC
        			DEFAULT 0.0
        			CHECK (randslip BETWEEN 0.0 AND 1.0),

                -- How long the tool that runs the test says it will take
        	duration        INTERVAL,

        	-- How often the test should be repeated.
        	repeat		INTERVAL,

        	-- Time after which scheduling stops.
        	until		TIMESTAMP WITH TIME ZONE
        			DEFAULT 'infinity',

        	-- Number of times successfully executed before scheduling stops
        	max_runs	NUMERIC,

        	-- Number of times the task has been scheduled for a
        	-- run.  Note that this only ever increases, even if
        	-- runs are deleted because of cancellation.
        	runs	  	NUMERIC
        			DEFAULT 0,

        	--
        	-- TESTING
        	--

        	-- Tool that will be used to run this test
        	tool	  	INTEGER
        			NOT NULL
        			REFERENCES tool(id)
        			ON DELETE CASCADE,

        	-- List of URIs for nodes participating in the test.
        	participants	JSONB,

        	-- Number of participants involved in the test, derived from
        	-- the participants field.
        	nparticipants	INTEGER,

        	-- This node's participant number
        	participant	INTEGER,

        	-- List of archives where the results are sent.
               	archives	JSONB,

        	--
        	-- MISCELLANEOUS
        	--

        	-- Whether or not the task should be scheduled
        	enabled	    	BOOLEAN
        			DEFAULT TRUE,

        	-- Hints used by the limit system
        	hints	    	JSONB
        );


        -- This should be used when someone looks up the external ID.  Bring
        -- the row ID a long so it can be pulled without having to consult the
        -- table.
        CREATE INDEX task_uuid
        ON task(uuid, id);

        -- This helps the 'json' query limiter in the REST API
        CREATE INDEX task_json
        ON task(json);

	t_version := t_version + 1;

    END IF;

    -- Version 1 to version 2
    IF t_version = 1
    THEN
	-- CLI equivalent to what's in the JSON task spec
        ALTER TABLE task ADD COLUMN
	cli JSON;

	-- Turn off triggers because the updates below bring out a bug
	-- in the old version.  Can't just turn off task_alter because
	-- it may not exist yet.

        ALTER TABLE task DISABLE TRIGGER USER;

	-- Calculate CLI for all existing tasks
	FOR v1_2_record IN (SELECT * FROM task)
	LOOP
	    v1_2_run_result := pscheduler_plugin_invoke(
                'test', v1_2_record.json #>> '{test, type}', 'spec-to-cli',
		v1_2_record.json #>> '{test, spec}' );
	    IF v1_2_run_result.status <> 0 THEN
	        RAISE NOTICE 'Unable to divine CLI from spec id %: %',
		  v1_2_run_result.status, v1_2_run_result.stderr;
		v1_2_run_result.stdout := '[ "(CLI Unavailable)" ]';
	    END IF;

	    UPDATE TASK SET cli = v1_2_run_result.stdout::JSON
            WHERE id = v1_2_record.id;
	END LOOP;

        ALTER TABLE task ENABLE TRIGGER USER;

	ALTER TABLE task
	ALTER COLUMN cli SET NOT NULL;

        t_version := t_version + 1;
    END IF;


    -- Version 2 to version 3
    -- Improve indexing on json field
    IF t_version = 2
    THEN

        DROP INDEX IF EXISTS task_json;

        -- This helps the 'json' query limiter in the REST API
        CREATE INDEX task_json
        ON task USING GIN (json);

        t_version := t_version + 1;
    END IF;


    -- Version 3 to version 4
    -- Adds indexes to aid cascading deletes
    IF t_version = 3
    THEN
        CREATE INDEX task_test ON task(test);
	CREATE INDEX task_tool ON task(tool);

        t_version := t_version + 1;
    END IF;

    -- Version 4 to version 5
    -- Adds 'post' column
    IF t_version = 4
    THEN
        -- Amount of time to wait before a result might be available
        ALTER TABLE task ADD COLUMN
        post INTERVAL DEFAULT 'P0';

        t_version := t_version + 1;
    END IF;

    -- Version 5 to version 6
    -- Add proper default to uuid column
    IF t_version = 5
    THEN
        ALTER TABLE task ALTER COLUMN uuid
        DROP NOT NULL;

        ALTER TABLE task ALTER COLUMN uuid
        SET DEFAULT gen_random_uuid();

        t_version := t_version + 1;
    END IF;

    -- Version 6 to version 7
    -- Add first_start column
    IF t_version = 6
    THEN
	-- When the first run of the task started
        ALTER TABLE task ADD COLUMN
	first_start TIMESTAMP WITH TIME ZONE;

        t_version := t_version + 1;
    END IF;

    -- Version 7 to version 8
    -- Add field to hold passed limits
    IF t_version = 7
    THEN
	ALTER TABLE task ADD COLUMN limits_passed JSON DEFAULT '[]'::JSON;

	UPDATE task SET limits_passed = DEFAULT WHERE limits_passed IS NULL;

        t_version := t_version + 1;
    END IF;


    -- Version 8 to version 9
    -- Remove deprecated randslip column
    IF t_version = 8
    THEN
	ALTER TABLE task DROP COLUMN randslip;

        t_version := t_version + 1;
    END IF;


    -- Version 9 to version 10
    -- Separate, queryable JSON column with details
    IF t_version = 9
    THEN
	ALTER TABLE task ADD COLUMN json_detail JSONB;

	-- Populate existing tasks
	UPDATE task SET json_detail = NULL;

        t_version := t_version + 1;
    END IF;


    -- Version 10 to version 11
    -- Add diagnostics column
    IF t_version = 10
    THEN
	ALTER TABLE task ADD COLUMN diags TEXT
	DEFAULT NULL;

	-- Add a placeholder for older tasks
	UPDATE task SET diags = '(None available)';

        t_version := t_version + 1;
    END IF;


    -- Version 11 to version 12
    -- Add runs_started column
    IF t_version = 11
    THEN
        -- Number of times a run has gone into the running state
	ALTER TABLE task ADD COLUMN runs_started NUMERIC
	DEFAULT 0;

        t_version := t_version + 1;
    END IF;

    -- Version 12 to version 13
    -- Add priority column.
    IF t_version = 12
    THEN
	ALTER TABLE task ADD COLUMN
	priority INTEGER;

        t_version := t_version + 1;
    END IF;

    -- Version 13 to version 14
    -- Adds cron_repeat column
    IF t_version = 13
    THEN
	-- How often the task should repeat, cron-style
	ALTER TABLE task ADD COLUMN
	repeat_cron TEXT;

        t_version := t_version + 1;
    END IF;


    -- Version 14 to version 15
    -- Adds participant_key column
    IF t_version = 14
    THEN

	-- Key to be used when tasking second participants when none
	-- was specified in the task.
	ALTER TABLE task ADD COLUMN
	participant_key TEXT;

        t_version := t_version + 1;
    END IF;


    -- Version 15 to version 16
    -- Drops limits_passed column
    IF t_version = 15
    THEN
	ALTER TABLE task DROP COLUMN limits_passed;

        t_version := t_version + 1;
    END IF;


    -- Version 16 to version 17
    -- Adds keep_after_archive column
    IF t_version = 16
    THEN
	ALTER TABLE task ADD COLUMN
	keep_after_archive INTERVAL DEFAULT NULL;

        t_version := t_version + 1;
    END IF;

    --
    -- Cleanup
    --

    PERFORM table_version_set(t_name, t_version, t_version_old);

END;
$$ LANGUAGE plpgsql;







DROP TRIGGER IF EXISTS task_alter ON task CASCADE;

DO $$ BEGIN PERFORM drop_function_all('task_alter'); END $$;

CREATE OR REPLACE FUNCTION task_alter()
RETURNS TRIGGER
AS $$
DECLARE
	key TEXT;
	test_type TEXT;
	tool_type TEXT;
	start TEXT;
	until TEXT;
	run_result external_program_result;
	temp_json JSONB;
	archive JSONB;
	json_result TEXT;
	context_participant JSONB;
	context JSONB;
        scheduling RECORD;
BEGIN

	--
	-- BASIC VALIDATION
	--

	-- TODO: Should probably only allow updates to the columns
	-- that we don't want changing, which is the task package.

	IF TG_OP = 'INSERT' THEN
	    NEW.added := now();
	ELSIF NEW.added <> OLD.added THEN
	    RAISE EXCEPTION 'Insertion time cannot be updated.';
	END IF;


	--
	-- TEST
	--

	IF (NEW.json #> '{test}') IS NULL THEN
	   RAISE EXCEPTION 'Task package has no test section';
	END IF;


	test_type := (NEW.json #>> '{test, type}');

	IF test_type IS NULL THEN
	   RAISE EXCEPTION 'Task contains no test type';
	END IF;

	SELECT INTO NEW.test id FROM test WHERE name = test_type;
	IF NOT FOUND THEN
            RAISE EXCEPTION 'Test type "%" is not available', test_type;
	END IF;


	-- Do this only if the JSON has changed
	IF TG_OP = 'INSERT' OR (TG_OP = 'UPDATE' AND NEW.json <> OLD.json)
        THEN

	    -- Calculate CLI equivalent
	    run_result := pscheduler_plugin_invoke('test', test_type, 'spec-to-cli',
		          NEW.json #>> '{test, spec}' );
	    IF run_result.status <> 0 THEN
	        RAISE EXCEPTION 'Unable to divine CLI from spec: %', run_result.stderr;
	    END IF;
	    NEW.cli = run_result.stdout::JSON;

	    -- Count the participants
	    NEW.nparticipants := jsonb_array_length(NEW.participants);
	    IF NEW.nparticipants IS NULL OR NEW.nparticipants = 0
            THEN
                RAISE EXCEPTION 'INTERNAL ERROR: Test produced empty participant list from task.';
            END IF;

	    -- Set the priority
	    IF (NEW.json #> '{priority}') IS NOT NULL THEN
	        NEW.priority := NEW.json #>> '{priority}';
	    END IF;

        END IF;

	-- TODO: Should probably check that the list members are
	-- unique and that one of them includes the local system.

	-- Validate the participant number
	IF NEW.participant IS NULL THEN
	    RAISE EXCEPTION 'A participant number must be provided.';
	END IF;
	IF (NEW.participant < 0)
	    OR (NEW.participant > (NEW.nparticipants - 1)) THEN
	    RAISE EXCEPTION 'Participant number is invalid for this test';
	END IF;

        --
        -- UUID
        --

	IF TG_OP = 'INSERT' THEN
            IF NEW.participant = 0 THEN
                IF NEW.uuid IS NOT NULL THEN
                    RAISE EXCEPTION 'Lead participant should not be given a UUID.';
                END IF;
                NEW.uuid = gen_random_uuid();
            ELSIF NEW.uuid IS NULL THEN
                RAISE EXCEPTION 'Non-lead participants must have a UUID.';
            END IF;
        END IF;


        --
        -- Key for second and later participants.  Use the one
        -- provided in the task if that exists.
        --

	IF TG_OP = 'INSERT' AND NEW.nparticipants > 1 AND new.participant = 0 THEN
            -- Create a key only if one wasn't provided.
            IF NEW.json ? '_key' THEN
	      NEW.participant_key := NEW.json ->> '_key';
	    ELSE
	      NEW.participant_key := random_string(64, TRUE, TRUE);
	    END IF;
        ELSIF TG_OP = 'UPDATE' AND NEW.participant_key <> OLD.participant_key THEN
            RAISE EXCEPTION 'Participant key cannot be updated.';
        END IF;

	--
	-- TOOL
	--

	tool_type := NEW.json #>> '{tool}';
	IF tool_type IS NULL THEN
	   RAISE EXCEPTION 'Task package specifies no tool';
	END IF;

	SELECT INTO NEW.tool id FROM tool WHERE name = tool_type AND available;

	IF NOT FOUND THEN
            RAISE EXCEPTION 'Tool "%" is not available', tool_type;
	END IF;



	--
	-- SCHEDULE
	--

	IF (NEW.json #> '{schedule}') IS NULL THEN
	    RAISE EXCEPTION 'Task package has no schedule section.';
	END IF;

	start := NEW.json #>> '{schedule, start}';
	IF start IS NULL or start = 'soonest' THEN
	    NEW.start := NULL;
        ELSIF start LIKE '@P%' THEN
            -- Start at the next increment of the specified interval,
            -- e.g., @PT1H means to start at the top of the next hour.
            -- Any interval not containing fractional seconds will
            -- work.
	    NEW.start := time_next_interval(now(),
                text_to_interval(substring(start FROM 2)));
	ELSIF start LIKE 'P%' THEN
	    -- ISO Interval, from right now.
	    NEW.start := normalized_time(now() + text_to_interval(start));
	ELSE
	    -- Full timestamp
	    NEW.start := normalized_time(text_to_timestamp_with_time_zone(start));
	END IF;


	NEW.slip := text_to_interval(NEW.json #>> '{schedule, slip}');
	IF NEW.slip IS NULL THEN
	   NEW.slip := 'P0';
	END IF;

	NEW.repeat := text_to_interval(NEW.json #>> '{schedule, repeat}');
	NEW.repeat_cron := NEW.json #>> '{schedule, repeat-cron}';
	IF NEW.repeat_cron IS NOT NULL AND NOT cron_spec_is_valid(NEW.repeat_cron) THEN
	    RAISE EXCEPTION 'Invalid cron repeat specification';
	END IF;

	-- TODO: Should we check that the repeat interval is greater
	-- than the duration (which we no longer have by default)?

	NEW.max_runs := text_to_numeric(NEW.json #>> '{schedule, max-runs}');
	IF (NEW.max_runs IS NOT NULL) AND (NEW.max_runs < 1) THEN
	   RAISE EXCEPTION 'Maximum runs must be positive.';
	END IF;


	-- REPEAT

	IF NEW.repeat IS NULL AND NEW.repeat_cron IS NULL THEN
	    -- Non-repeaters run once
	    NEW.until := NULL;
	    NEW.max_runs := 1;
	END IF;


	-- UNTIL

	until := NEW.json #>> '{schedule, until}';

	-- TODO: Make this compatible with TimestampAbsoluteRelative
	IF until IN ('forever', 'infinity', 'doomsday') THEN
	   NEW.until := 'infinity';
	ELSE
	   NEW.until := text_to_timestamp_with_time_zone(until);
	END IF;

	IF ( (TG_OP = 'INSERT')
             OR (TG_OP = 'UPDATE' AND NEW.until <> OLD.until) )
           AND NEW.until IS NOT NULL
	   AND NEW.until <= now() THEN
	    RAISE EXCEPTION 'Until must be in the future.';
        END IF;

        IF NEW.start IS NOT NULL AND NEW.until <= NEW.start THEN
	   RAISE EXCEPTION 'Until must be after the start.';
	END IF;


        -- See what the tool says about how long it should take.
	
	IF TG_OP = 'INSERT' OR (TG_OP = 'UPDATE' AND NEW.json <> OLD.json)
        THEN
	    run_result := pscheduler_plugin_invoke('tool', tool_type, 'duration',
		          NEW.json #>> '{test}' );
	    IF run_result.status <> 0 THEN
	        RAISE EXCEPTION 'Unable to determine duration of test: %', run_result.stderr;
	    END IF;
	    temp_json := run_result.stdout::JSONB;
	    NEW.duration := interval_round_up( (temp_json #>> '{duration}')::INTERVAL );
	    -- This will come up NULL if not provided.
	    NEW.post := interval_round_up( (temp_json #>> '{post}')::INTERVAL );
	    IF NEW.post IS NULL THEN
                NEW.post = 'P0'::INTERVAL;
            END IF;
        END IF;

	--
	-- ARCHIVES
	--

	IF (TG_OP = 'INSERT'
            OR (TG_OP = 'UPDATE' AND NEW.json <> OLD.json))
	   AND NEW.json ? 'archives'
        THEN
	    FOR archive IN (SELECT * FROM jsonb_array_elements_text(NEW.json -> 'archives'))
	    LOOP
	        PERFORM archiver_validate(archive);
	    END LOOP;

	    -- Only needed if archives are specified.
	    NEW.keep_after_archive := text_to_interval(NEW.json #>> '{keep-after-archive}');
	END IF;



	--
	-- CONTEXTS
	--

	IF (TG_OP = 'INSERT'
            OR (TG_OP = 'UPDATE' AND NEW.json <> OLD.json))
	   AND NEW.json ? 'contexts'
        THEN

	    -- Make sure it looks sane
	    json_result := json_validate(NEW.json #> '{contexts}', '#/pScheduler/ContextSpecification');
            IF json_result IS NOT NULL
            THEN
		RAISE EXCEPTION 'Invalid context specification: %', json_result;
	    END IF;

	    -- Make sure there aren't more contexts than the task has participants.
	    IF jsonb_array_length(NEW.json #> '{contexts,contexts}') <> NEW.nparticipants
            THEN
		RAISE EXCEPTION 'Number of contexts for this task must be exactly %',
		    NEW.nparticipants;
	    END IF;

	    -- Check with all of the plugins to make sure the data is valid.
	    FOR context_participant IN ( SELECT * FROM jsonb_array_elements(NEW.json #> '{contexts, contexts}') )
	    LOOP
	        FOR context IN (SELECT * FROM jsonb_array_elements_text(context_participant))
		LOOP
		    PERFORM context_validate(context);
	        END LOOP;
	    END LOOP;
	END IF;

	--
	-- SEARCHABLE JSON WITH DETAILS
	--

	SELECT INTO scheduling
            scheduling_class.*
        FROM
	    test
            JOIN scheduling_class ON scheduling_class.id = test.scheduling_class
        WHERE
            test.id = NEW.test
        ;


	NEW.json_detail := jsonb_set(NEW.json, '{detail}', '{}'::JSONB);

	NEW.json_detail := jsonb_set(NEW.json_detail, '{detail,added}',
            to_jsonb(timestamp_with_time_zone_to_iso8601(NEW.added)));

	NEW.json_detail := jsonb_set(NEW.json_detail, '{detail,anytime}',
            to_jsonb(scheduling.anytime));

	NEW.json_detail := jsonb_set(NEW.json_detail, '{detail,cli}',
            to_jsonb(NEW.cli));

	NEW.json_detail := jsonb_set(NEW.json_detail, '{detail,diags}',
            to_jsonb(NEW.diags));

	NEW.json_detail := jsonb_set(NEW.json_detail, '{detail,duration}',
            to_jsonb(interval_to_iso8601(NEW.duration)));

	NEW.json_detail := jsonb_set(NEW.json_detail, '{detail,enabled}',
	    to_jsonb(NEW.enabled));

	NEW.json_detail := jsonb_set(NEW.json_detail, '{detail,exclusive}',
            to_jsonb(scheduling.exclusive));

	NEW.json_detail := jsonb_set(NEW.json_detail, '{detail,hints}',
            to_jsonb(NEW.hints));

	NEW.json_detail := jsonb_set(NEW.json_detail, '{detail,multi-result}',
            to_jsonb(scheduling.multi_result));

	NEW.json_detail := jsonb_set(NEW.json_detail, '{detail,participant}',
	    to_jsonb(NEW.participant));

	NEW.json_detail := jsonb_set(NEW.json_detail, '{detail,participants}',
	    to_jsonb(NEW.participants));

	NEW.json_detail := jsonb_set(NEW.json_detail, '{detail,post}',
            to_jsonb(interval_to_iso8601(NEW.post)));

	-- TODO: runs and runs_started change often enough that it
	-- might be worth hitting those in separate triggers.

	NEW.json_detail := jsonb_set(NEW.json_detail, '{detail,runs}',
	    to_jsonb(NEW.runs));

	NEW.json_detail := jsonb_set(NEW.json_detail, '{detail,runs-started}',
	    to_jsonb(NEW.runs_started));

        IF NEW.slip IS NOT NULL
        THEN
	    NEW.json_detail := jsonb_set(NEW.json_detail, '{detail,slip}',
                to_jsonb(interval_to_iso8601(NEW.slip)));
        ELSE
            NEW.json_detail := jsonb_set(NEW.json_detail, '{detail,slip}',
                'null'::JSONB);
        END IF;

        IF NEW.start IS NOT NULL
        THEN
	    NEW.json_detail := jsonb_set(NEW.json_detail, '{detail,start}',
                to_jsonb(timestamp_with_time_zone_to_iso8601(NEW.start)));
        ELSE
            NEW.json_detail := jsonb_set(NEW.json_detail, '{detail,start}',
                'null'::JSONB);
        END IF;

	RETURN NEW;
END;
$$ LANGUAGE plpgsql;

CREATE TRIGGER task_alter BEFORE INSERT OR UPDATE ON task
    FOR EACH ROW EXECUTE PROCEDURE task_alter();




DROP TRIGGER IF EXISTS task_alter_notify ON task CASCADE;

DO $$ BEGIN PERFORM drop_function_all('task_alter_notify'); END $$;

CREATE OR REPLACE FUNCTION task_alter_notify()
RETURNS TRIGGER
AS $$
BEGIN
    NOTIFY task_change;
    RETURN NEW;
END;
$$ LANGUAGE plpgsql;

CREATE TRIGGER task_alter_notify AFTER INSERT ON task
    -- Doesn't matter how many inserts happened
    FOR EACH STATEMENT
    EXECUTE PROCEDURE task_alter_notify();

CREATE TRIGGER task_alter_notify_update AFTER UPDATE ON task
    -- One of these per row since changes may be different.
    FOR EACH ROW
    -- Don't trigger scheduling if it's just a run count update.
    WHEN ((NEW.runs = OLD.runs) AND (NEW.runs_started = OLD.runs_started))
    EXECUTE PROCEDURE task_alter_notify();




-- Calculate when the next iteration of a task should take place after
-- a given time.  This isn't actually a task-specific function, but
-- tasks are the only context where it will be used.

DO $$ BEGIN PERFORM drop_function_all('task_next_run'); END $$;

CREATE OR REPLACE FUNCTION task_next_run(
       start TIMESTAMP WITH TIME ZONE,   -- Task's start (first run) time
       after TIMESTAMP WITH TIME ZONE,   -- Base for "next" time
       repeat INTERVAL,                  -- Time between runs
       repeat_cron TEXT                  -- Cron-style repeat spec
)
RETURNS TIMESTAMP WITH TIME ZONE
AS $$
DECLARE
    intervals NUMERIC;
    next_interval TIMESTAMP WITH TIME ZONE;
    next_cron TIMESTAMP WITH TIME ZONE;
BEGIN

    IF repeat IS NULL AND repeat_cron IS NULL THEN
        RAISE EXCEPTION 'Must have either an interval or a cron spec';
    END IF;

    -- Standard, every-x repeat

    IF repeat IS NULL THEN
        next_interval := tstz_infinity();
    ELSIF repeat = 'P0' THEN
        RAISE EXCEPTION 'Cannot divide by a zero interval';
    ELSE 
        -- Number of runs that should have happened between the task
        -- start and the after time.
	intervals := 
        	  TRUNC( (EXTRACT(EPOCH FROM after) - EXTRACT(EPOCH FROM start))
		  / EXTRACT(EPOCH FROM repeat) ) + 1;
	next_interval := start + (repeat * intervals);
    END IF;

    -- Cron-style repeat

    IF repeat_cron IS NULL THEN
        next_cron = tstz_infinity();
    ELSE
        next_cron = cron_next(repeat_cron, after);
    END IF;

    -- The behavior in the presence of both is to use the earlier
    RETURN least(next_interval, next_cron);

END;
$$ LANGUAGE plpgsql;





---
--- Maintenance
---

DO $$ BEGIN PERFORM drop_function_all('task_purge'); END $$;

CREATE OR REPLACE FUNCTION task_purge()
RETURNS VOID
AS $$
DECLARE
    older_than TIMESTAMP WITH TIME ZONE;
BEGIN

    SELECT INTO older_than normalized_now() - keep_runs_tasks
    FROM configurables;

    -- Get rid of tasks that no longer have runs and can be considered
    -- completed.

    DELETE FROM task
    WHERE
        -- TODO: This refers to a table created later.
        NOT EXISTS (SELECT * FROM run where run.task = task.id)
        -- The first of these will be the latest known time.
        AND COALESCE(until, start, added) < older_than
        AND (
            -- Complete based on runs
            (max_runs IS NOT NULL AND runs >= max_runs)
            -- One-shot
            OR (repeat IS NULL AND repeat_cron IS NULL)
            -- Until time has passed
            OR until < older_than
            )
    ;

END;
$$ LANGUAGE plpgsql;



-- Maintenance that happens four times a minute.

DO $$ BEGIN PERFORM drop_function_all('task_maint_fifteen'); END $$;

CREATE OR REPLACE FUNCTION task_maint_fifteen()
RETURNS VOID
AS $$
BEGIN
    PERFORM task_purge();
END;
$$ LANGUAGE plpgsql;



---
--- API
---


-- Put a task on the timeline and return its UUID

DO $$ BEGIN PERFORM drop_function_all('api_task_post'); END $$;

CREATE OR REPLACE FUNCTION api_task_post(
    task_package JSONB,
    participant_list TEXT[],
    hints JSONB,
    participant INTEGER DEFAULT 0,
    priority INTEGER DEFAULT NULL,
    task_uuid UUID = NULL,
    enabled BOOLEAN = TRUE,
    diags TEXT = '(None)'
)
RETURNS TABLE(
  uuid UUID,
  participant_key TEXT
)
AS $$
DECLARE
    inserted RECORD;
BEGIN

   IF EXISTS (SELECT * FROM task WHERE task.uuid = task_uuid)
   THEN
       RAISE EXCEPTION 'Task already exists.  All participants must be on separate systems.';
   END IF;

   WITH inserted_row AS (
        INSERT INTO task(json, participants, participant,
	                 priority, uuid, hints, enabled, diags)
        VALUES (task_package, array_to_json(participant_list),
	        participant, priority, task_uuid, hints, enabled, diags)
        RETURNING *
    ) SELECT INTO inserted * from inserted_row;

    uuid := inserted.uuid;
    participant_key := inserted.participant_key;

    RETURN NEXT;

END;
$$ LANGUAGE plpgsql;



-- This function enables a task by its UUID.  This is used by the REST
-- API to keep the scheduler from trying to ask the other participants
-- for runtimes when they don't have it yet.  (Other participants'
-- schedulers won't touch their copies because they're not leading.)

DO $$ BEGIN PERFORM drop_function_all('api_task_enable'); END $$;

CREATE OR REPLACE FUNCTION api_task_enable(
    task_uuid UUID        -- UUID of task to enable
)
RETURNS VOID
AS $$
DECLARE
    taskrec RECORD;
BEGIN

    SELECT INTO taskrec * FROM task WHERE uuid = task_uuid;
    IF NOT FOUND
    THEN
        RAISE EXCEPTION 'Task not found while enabling task %', task_uuid;
    END IF;

    IF taskrec.enabled
    THEN
        -- Don't so anything redundant redundant.
        RETURN;
    END IF;

    UPDATE task SET enabled = TRUE WHERE id = taskrec.id;

END;
$$ LANGUAGE plpgsql;


-- This function disables a task and does the corresponding DELETE on
-- the other participants.  The task_url_format argument is the URL of
-- the task with the hostname replaced with %s, which will be filled
-- in with the hostname of each participant.

DO $$ BEGIN PERFORM drop_function_all('api_task_disable'); END $$;

CREATE OR REPLACE FUNCTION api_task_disable(
    task_uuid UUID,       -- UUID of task to disable
    task_url_format TEXT  -- URL template.  See above.
)
RETURNS VOID
AS $$
DECLARE
    taskrec RECORD;
    host TEXT;
    ip INET;
    ip_family INTEGER;
    bind TEXT;
    task_url_append TEXT;
BEGIN

    SELECT INTO taskrec * FROM task WHERE uuid = task_uuid;
    IF NOT FOUND
    THEN
        RAISE EXCEPTION 'Task not found while disabling task %', task_uuid;
    END IF;

    IF NOT taskrec.enabled
    THEN
        -- Don't so anything redundant redundant.
        RETURN;
    END IF;

    UPDATE task SET enabled = FALSE WHERE uuid = task_uuid;

    -- Sling a DELETE at the non-lead participants if there are any
    -- and we're the lead.

    IF taskrec.participant = 0
    THEN

        -- If the task has a lead-bind, use it.
	IF taskrec.json ? 'lead-bind'
	THEN
	    bind := taskrec.json #>> '{"lead-bind"}';
	ELSE
	    bind := NULL;
	END IF;

	-- If this task has a key, use that for the participant key or
	-- generate one.

        IF taskrec.participant_key IS NOT NULL
	THEN
	    task_url_append := '?key=' || uri_encode(taskrec.participant_key);
	ELSE
	    task_url_append := '';
	END IF;


        FOR host IN (SELECT jsonb_array_elements_text(taskrec.participants)
                     FROM task WHERE uuid = task_uuid OFFSET 1)
        LOOP

            -- IPv6 adresses get special treatment
            BEGIN
		IF family(host::INET) = 6
                THEN
                    host := format('[%s]', host);
                END IF;
            EXCEPTION WHEN OTHERS THEN
                NULL;  -- Don't care
            END;

            INSERT INTO http_queue (operation, uri, bind)
                VALUES ('DELETE',
		    format(task_url_format, host) || task_url_append,
		    bind);
        END LOOP;
    END IF;   

END;
$$ LANGUAGE plpgsql;
--
-- Run State
--

DO $$
DECLARE
    t_name TEXT;            -- Name of the table being worked on
    t_version INTEGER;      -- Current version of the table
    t_version_old INTEGER;  -- Version of the table at the start
BEGIN

    --
    -- Preparation
    --

    t_name := 'run_state';

    t_version := table_version_find(t_name);
    t_version_old := t_version;


    --
    -- Upgrade Blocks
    --

    -- Version 0 (nonexistant) to version 1
    IF t_version = 0
    THEN

        CREATE TABLE run_state (

        	-- Row identifier
        	id		INTEGER
        			PRIMARY KEY,

        	-- Display name
        	display		TEXT
        			UNIQUE NOT NULL,

        	-- Enumeration for use by programs
        	enum		TEXT
        			UNIQUE NOT NULL
        );

	t_version := t_version + 1;

    END IF;

    -- Version 1 to version 2
    IF t_version = 1
    THEN

        -- This column indicates if there's nothing more to be done
        -- with this run, successful or not.
        ALTER TABLE run_state ADD COLUMN
        finished BOOLEAN DEFAULT FALSE;

        t_version := t_version + 1;
    END IF;


    -- Version 2 to version 3
    IF t_version = 2
    THEN

        -- Indicates whether the state indicates success (TRUE),
        -- failure (FALSE) or neither (NULL).
        ALTER TABLE run_state ADD COLUMN
        success BOOLEAN DEFAULT NULL;

        t_version := t_version + 1;
    END IF;


    --
    -- Cleanup
    --

    PERFORM table_version_set(t_name, t_version, t_version_old);

END;
$$ LANGUAGE plpgsql;



--
-- Functions that encapsulate the numeric values for each state
--

-- NOTE: These use ALTER FUNCTION to set the IMMUTABLE attribute
-- because in some cases, replacement doesn't set it properly.


-- Run is being scheduled and should be invisible to the API

DO $$ BEGIN PERFORM drop_function_all('run_state_scheduling'); END $$;

CREATE OR REPLACE FUNCTION run_state_scheduling()
RETURNS INTEGER
AS $$
BEGIN
	RETURN 0;
END;
$$ LANGUAGE plpgsql;
ALTER FUNCTION run_state_scheduling() IMMUTABLE;


-- Run is waiting to execute (not time yet)

DO $$ BEGIN PERFORM drop_function_all('run_state_pending'); END $$;

CREATE OR REPLACE FUNCTION run_state_pending()
RETURNS INTEGER
AS $$
BEGIN
	RETURN 1;
END;
$$ LANGUAGE plpgsql;
ALTER FUNCTION run_state_pending() IMMUTABLE;


-- The runner is preparing to execute the run

DO $$ BEGIN PERFORM drop_function_all('run_state_on_deck'); END $$;

CREATE OR REPLACE FUNCTION run_state_on_deck()
RETURNS INTEGER
AS $$
BEGIN
	RETURN 2;
END;
$$ LANGUAGE plpgsql;
ALTER FUNCTION run_state_on_deck() IMMUTABLE;


-- Run is being executed

DO $$ BEGIN PERFORM drop_function_all('run_state_running'); END $$;

CREATE OR REPLACE FUNCTION run_state_running()
RETURNS INTEGER
AS $$
BEGIN
	RETURN 3;
END;
$$ LANGUAGE plpgsql;
ALTER FUNCTION run_state_running() IMMUTABLE;

-- Post-run cleanup

DO $$ BEGIN PERFORM drop_function_all('run_state_cleanup'); END $$;

CREATE OR REPLACE FUNCTION run_state_cleanup()
RETURNS INTEGER
AS $$
BEGIN
	RETURN 4;
END;
$$ LANGUAGE plpgsql;
ALTER FUNCTION run_state_cleanup() IMMUTABLE;


-- Run finished successfully

DO $$ BEGIN PERFORM drop_function_all('run_state_finished'); END $$;

CREATE OR REPLACE FUNCTION run_state_finished()
RETURNS INTEGER
AS $$
BEGIN
	RETURN 5;
END;
$$ LANGUAGE plpgsql;
ALTER FUNCTION run_state_finished() IMMUTABLE;


-- No idea of the outcome yet

DO $$ BEGIN PERFORM drop_function_all('run_state_overdue'); END $$;

CREATE OR REPLACE FUNCTION run_state_overdue()
RETURNS INTEGER
AS $$
BEGIN
	RETURN 6;
END;
$$ LANGUAGE plpgsql;
ALTER FUNCTION run_state_overdue() IMMUTABLE;


-- Run never happened

DO $$ BEGIN PERFORM drop_function_all('run_state_missed'); END $$;

CREATE OR REPLACE FUNCTION run_state_missed()
RETURNS INTEGER
AS $$
BEGIN
	RETURN 7;
END;
$$ LANGUAGE plpgsql;
ALTER FUNCTION run_state_missed() IMMUTABLE;


-- Run ran but was not a success

DO $$ BEGIN PERFORM drop_function_all('run_state_failed'); END $$;

CREATE OR REPLACE FUNCTION run_state_failed()
RETURNS INTEGER
AS $$
BEGIN
	RETURN 8;
END;
$$ LANGUAGE plpgsql;
ALTER FUNCTION run_state_failed() IMMUTABLE;


-- Run lost out to something with a higher priority

DO $$ BEGIN PERFORM drop_function_all('run_state_preempted'); END $$;

CREATE OR REPLACE FUNCTION run_state_preempted()
RETURNS INTEGER
AS $$
BEGIN
	RETURN 9;
END;
$$ LANGUAGE plpgsql;
ALTER FUNCTION run_state_preempted() IMMUTABLE;

-- Run was dead on arrival

DO $$ BEGIN PERFORM drop_function_all('run_state_nonstart'); END $$;

CREATE OR REPLACE FUNCTION run_state_nonstart()
RETURNS INTEGER
AS $$
BEGIN
	RETURN 10;
END;
$$ LANGUAGE plpgsql;
ALTER FUNCTION run_state_nonstart() IMMUTABLE;


-- Run was canceled

DO $$ BEGIN PERFORM drop_function_all('run_state_canceled'); END $$;

CREATE OR REPLACE FUNCTION run_state_canceled()
RETURNS INTEGER
AS $$
BEGIN
	RETURN 11;
END;
$$ LANGUAGE plpgsql;
ALTER FUNCTION run_state_canceled() IMMUTABLE;



DROP TRIGGER IF EXISTS run_state_alter ON run_state CASCADE;

DO $$ BEGIN PERFORM drop_function_all('run_state_alter'); END $$;

CREATE OR REPLACE FUNCTION run_state_alter()
RETURNS TRIGGER
AS $$
BEGIN
	RAISE EXCEPTION 'This table may not be altered';
	RETURN NEW;
END;
$$ LANGUAGE plpgsql;

CREATE TRIGGER run_state_alter
BEFORE INSERT OR UPDATE OR DELETE
ON run_state
       FOR EACH ROW EXECUTE PROCEDURE run_state_alter();



-- Do this after trigger creation and with a DISABLE/ENABLE in case
-- the table was previously populated.

ALTER TABLE run_state DISABLE TRIGGER run_state_alter;
INSERT INTO run_state (id, display, enum, finished, success)
VALUES
    (run_state_scheduling(),'Scheduling',  'scheduling',FALSE, NULL),
    (run_state_pending(),   'Pending',     'pending',   FALSE, NULL),
    (run_state_on_deck(),   'On Deck',     'on-deck',   FALSE, NULL),
    (run_state_running(),   'Running',     'running',   FALSE, NULL),
    (run_state_cleanup(),   'Cleanup',     'cleanup',   TRUE,  NULL),
    (run_state_finished(),  'Finished',    'finished',  TRUE,  TRUE),
    (run_state_overdue(),   'Overdue',     'overdue',   TRUE,  FALSE),
    (run_state_missed(),    'Missed',      'missed',    TRUE,  FALSE),
    (run_state_failed(),    'Failed',      'failed',    TRUE,  FALSE),
    (run_state_preempted(), 'Preempted',   'preempted', TRUE,  FALSE),
    (run_state_nonstart(),  'Non-Starter', 'nonstart',  TRUE,  FALSE),
    (run_state_canceled(),  'Canceled',    'canceled',  TRUE,  NULL)
ON CONFLICT (id) DO UPDATE
SET
    display = EXCLUDED.display,
    enum = EXCLUDED.enum,
    finished = EXCLUDED.finished,
    success = EXCLUDED.success;
ALTER TABLE run_state ENABLE TRIGGER run_state_alter;


-- Determine if a transition between states is valid
DO $$ BEGIN PERFORM drop_function_all('run_state_transition_is_valid'); END $$;

CREATE OR REPLACE FUNCTION run_state_transition_is_valid(
    old INTEGER,
    new INTEGER
)
RETURNS BOOLEAN
AS $$
BEGIN
   -- TODO: This might be worth putting into a table.
   RETURN  new = old
           OR   ( old = run_state_scheduling()
	          AND new IN (run_state_pending(),
			      run_state_preempted(),
			      run_state_canceled(),
			      run_state_nonstart()) )
           OR   ( old = run_state_pending()
	          AND new IN (run_state_on_deck(),
			      run_state_running(),
			      run_state_missed(),
			      run_state_failed(),
			      run_state_preempted(),
			      run_state_canceled(),
			      run_state_nonstart()) )
           OR   ( old = run_state_on_deck()
	          AND new IN (run_state_running(),
		              run_state_overdue(),
			      run_state_missed(),
			      run_state_failed(),
			      run_state_canceled(),
			      run_state_preempted()) )
           OR ( old = run_state_running()
	        AND new IN (run_state_cleanup(),
		            run_state_finished(),
		            run_state_overdue(),
			    run_state_missed(),
			    run_state_failed(),
			    run_state_preempted(),
			    run_state_canceled()) )
           OR ( old = run_state_cleanup()
	        AND new IN (run_state_finished(),
			    run_state_failed()) )
           OR ( old = run_state_finished()
	        AND new IN (run_state_failed()) )
	   OR ( old = run_state_overdue()
                AND new IN (run_state_finished(),
		            run_state_missed(),
		            run_state_failed(),
			    run_state_canceled(),
		            run_state_preempted()) )
           OR ( old = run_state_missed()
	        AND new IN (run_state_failed()) )
           OR   ( old = run_state_nonstart()
	          AND new IN (run_state_canceled()) )
           ;
END;
$$ LANGUAGE plpgsql;





-- Determine if a run state is one of the finished ones.

DO $$ BEGIN PERFORM drop_function_all('run_state_is_finished'); END $$;

CREATE OR REPLACE FUNCTION run_state_is_finished(
    state INTEGER
)
RETURNS BOOLEAN
AS $$
BEGIN
    RETURN EXISTS (SELECT * FROM run_state WHERE id = state AND finished);
END;
$$ LANGUAGE plpgsql
IMMUTABLE;
--
-- Run Table
--

DO $$
DECLARE
    t_name TEXT;            -- Name of the table being worked on
    t_version INTEGER;      -- Current version of the table
    t_version_old INTEGER;  -- Version of the table at the start
BEGIN

    --
    -- Preparation
    --

    t_name := 'run';

    t_version := table_version_find(t_name);
    t_version_old := t_version;


    --
    -- Upgrade Blocks
    --

    -- Version 0 (nonexistant) to version 1
    IF t_version = 0
    THEN

        CREATE TABLE run (

        	-- Row identifier
        	id		BIGSERIAL
        			PRIMARY KEY,

        	-- External-use identifier
        	uuid		UUID
        			UNIQUE
        			DEFAULT gen_random_uuid(),

        	-- Task this run belongs to
        	task		BIGINT
        			REFERENCES task(id)
        			ON DELETE CASCADE,

        	-- Range of times when this task will be run
        	times		TSTZRANGE
        			NOT NULL,

        	--
        	-- Information about the local system's participation in the
        	-- test
        	--

                -- Participant data for the local test
                part_data        JSONB,

        	-- State of this run
        	state	    	 INTEGER DEFAULT run_state_pending()
        			 REFERENCES run_state(id),

        	-- Any errors that prevented the run from being put on the
        	-- schedule, used when state is run_state_nonstart().  Any
        	-- test- or tool-related errors will be incorporated into the
        	-- local or merged results.
        	errors   	 TEXT,

        	-- How it went locally, i.e., what the test returned
        	-- TODO: See if this is used anywhere.
        	status           INTEGER,

        	-- Result from the local run
        	-- TODO: Change this to local_result to prevent confusion
        	result   	 JSONB,

        	--
        	-- Information about the whole test
        	--

                -- Participant data for all participants in the test.  This is
                -- an array, with each element being the part_data for
                -- each participant.
                part_data_full   JSONB,

        	-- Combined resut generated by the lead participant
        	result_full    	 JSONB,

        	-- Merged result generated by the tool that did the test
        	result_merged  	 JSONB,

        	-- Clock survey, done if the run was not successful.
        	clock_survey  	 JSONB
        );

        -- This should be used when someone looks up the external ID.  Bring
        -- the row ID a long so it can be pulled without having to consult the
        -- table.
        CREATE INDEX run_uuid ON run(uuid, id);

        -- GIST accelerates range-specific operators like &&
        CREATE INDEX run_times ON run USING GIST (times);

        -- These two indexes are used by the schedule_gap view.
        CREATE INDEX run_times_lower ON run(lower(times), state);
        CREATE INDEX run_times_upper ON run(upper(times));

	t_version := t_version + 1;

    END IF;

    -- Version 1 to version 2
    -- Adds indexes for task to aid cascading deletes
    IF t_version = 1
    THEN
        CREATE INDEX run_task ON run(task);

        t_version := t_version + 1;
    END IF;

    -- Version 2 to version 3
    -- Adds index for upcoming/current runs
    IF t_version = 2
    THEN
        CREATE INDEX run_current ON run(state)
        WHERE state in (
            run_state_pending(),
            run_state_on_deck(),
            run_state_running()
        );

        t_version := t_version + 1;
    END IF;

    -- Version 3 to version 4
    -- Rebuilds index from previous version, which didn't have
    -- IMMUTABLE versions of the run_state_* functions.
    IF t_version = 3
    THEN
        DROP INDEX IF EXISTS run_current CASCADE;
        CREATE INDEX run_current ON run(state)
        WHERE state in (
            run_state_pending(),
            run_state_on_deck(),
            run_state_running()
        );

        t_version := t_version + 1;
    END IF;

    -- Version 4 to version 5
    -- Adds state to index of upper times
    IF t_version = 4
    THEN
        DROP INDEX IF EXISTS run_times_upper;
        CREATE INDEX run_times_upper ON run(upper(times), state);

        t_version := t_version + 1;
    END IF;


    -- Version 5 to version 6
    -- Adds 'added' column
    IF t_version = 5
    THEN
        ALTER TABLE run ADD COLUMN
        added TIMESTAMP WITH TIME ZONE;

        t_version := t_version + 1;
    END IF;


    -- Version 6 to version 7
    -- Adds 'times_actual' column
    IF t_version = 6
    THEN
        ALTER TABLE run ADD COLUMN
        times_actual TSTZRANGE NULL;

        t_version := t_version + 1;
    END IF;


    -- Version 7 to version 8
    -- Adds 'priority' and 'limit_diags' columns
    IF t_version = 7
    THEN
        ALTER TABLE run ADD COLUMN
        priority INTEGER DEFAULT 0;

        ALTER TABLE run ADD COLUMN
        limit_diags TEXT;

        t_version := t_version + 1;
    END IF;


    -- Version 8 to version 9
    -- New rows start in run_state_scheduling()
    IF t_version = 8
    THEN
        ALTER TABLE run
        ALTER COLUMN state
	SET DEFAULT run_state_scheduling();

        t_version := t_version + 1;
    END IF;


    -- Version 9 to version 10
    -- Add expiration time
    IF t_version = 9
    THEN
        ALTER TABLE run ADD COLUMN
	expires TIMESTAMP WITH TIME ZONE DEFAULT NULL;

	CREATE INDEX run_expires ON run(expires) WHERE expires IS NOT NULL;

        t_version := t_version + 1;
    END IF;


    --
    -- Cleanup
    --

    PERFORM table_version_set(t_name, t_version, t_version_old);

END;
$$ LANGUAGE plpgsql;





-- Runs which could cause conflicts

DROP VIEW IF EXISTS run_conflictable CASCADE;
CREATE OR REPLACE VIEW run_conflictable
AS
    SELECT
        run.*,
        task.duration,
        scheduling_class.anytime,
        scheduling_class.exclusive
    FROM
        run
        JOIN task ON task.id = run.task
	JOIN test ON test.id = task.test
        JOIN scheduling_class ON scheduling_class.id = test.scheduling_class
    WHERE
        run.state <> run_state_nonstart()
        AND NOT scheduling_class.anytime
;



-- Determine if a proposed run would have conflicts

DO $$ BEGIN PERFORM drop_function_all('run_has_conflicts'); END $$;

CREATE OR REPLACE FUNCTION run_has_conflicts(
    task_id BIGINT,
    proposed_start TIMESTAMP WITH TIME ZONE,
    proposed_priority INTEGER = NULL
)
RETURNS BOOLEAN
AS $$
DECLARE
    taskrec RECORD;
    proposed_times TSTZRANGE;
    start_margin INTERVAL;
BEGIN

    SELECT INTO taskrec
        task.*,
        test.scheduling_class,
        scheduling_class.anytime,
        scheduling_class.exclusive
    FROM
        task
        JOIN test ON test.id = task.test
        JOIN scheduling_class ON scheduling_class.id = test.scheduling_class
    WHERE
        task.id = task_id;

    IF NOT FOUND
    THEN
        RAISE EXCEPTION 'No such task.';
    END IF;

    -- Anytime tasks don't ever count
    IF taskrec.anytime
    THEN
        RETURN FALSE;
    END IF;

    proposed_times := tstzrange(proposed_start,
        proposed_start + taskrec.duration, '[)');

    SELECT INTO start_margin run_start_margin FROM configurables;

    RETURN ( 
        -- Exclusive can't collide with anything
        ( taskrec.exclusive
          AND EXISTS (SELECT * FROM run_conflictable
                      WHERE
		          times && proposed_times
			  AND COALESCE(priority, 0) >= COALESCE(proposed_priority, 0)
		     )
	)

        -- Non-exclusive can't collide with exclusive
          OR ( NOT taskrec.exclusive
               AND EXISTS (SELECT * FROM run_conflictable
                           WHERE
			       exclusive
			       AND times && proposed_times
			       AND COALESCE(priority, 0) >= COALESCE(proposed_priority, 0)
		    )
	    )

        -- Every run is kicked off a few seconds (see run_start_margin
        -- in the configurables table) before its scheduled start time
        -- so it can do advance prep and start the measurement at
        -- exactly the right time (see #970).  The tool plugin cannot
        -- handle being preempted during the start margin, so anything
        -- that would preempt it cannot be inserted during that
        -- period.  Beforehand is okay because the runner will see the
        -- preemption and not start the plugin.

	-- TODO: This could *probably* be folded into the two queries
	-- above, but it might get hairy.

	OR ( EXISTS (SELECT * FROM run_conflictable
                     WHERE
		         -- Proposed run overlaps an existing conflictable task
			 times && proposed_times
			 -- Proposed run would preempt existing run
			 AND COALESCE(proposed_priority, 0) >= COALESCE(priority, 0)
			 -- Proposed and existing runs can't coexist
			 AND ( exclusive OR (taskrec.exclusive AND NOT exclusive) )
			 -- Proposal is being made within the existing run's start margin
			 AND now() BETWEEN lower(times) - start_margin AND lower(times)
			 -- Proposed run would start within the existing's start margin
			 AND proposed_start BETWEEN lower(times) - start_margin AND lower(times)
	         )
	   )

    );

END;
$$ LANGUAGE plpgsql;





-- Standard message used when throwing conflict exceptions.  Used by
-- api_run_post() in a comparison.

DO $$ BEGIN PERFORM drop_function_all('run_conflict_message'); END $$;

CREATE OR REPLACE FUNCTION run_conflict_message()
RETURNS TEXT
AS $$
BEGIN
    RETURN 'Run would create a scheduling conflict.';
END;
$$ LANGUAGE plpgsql;




DROP TRIGGER IF EXISTS run_alter ON run CASCADE;

DO $$ BEGIN PERFORM drop_function_all('run_alter'); END $$;

CREATE OR REPLACE FUNCTION run_alter()
RETURNS TRIGGER
AS $$
DECLARE
    horizon INTERVAL;
    taskrec RECORD;
    tool_name TEXT;
    run_result external_program_result;
    pdata_out JSONB;
    result_merge_input JSONB;
BEGIN

    -- TODO: What changes to a run don't we allow?

    SELECT INTO taskrec
        task.*,
        test.scheduling_class,
        scheduling_class.anytime,
        scheduling_class.exclusive
    FROM
        task
        JOIN test ON test.id = task.test
        JOIN scheduling_class ON scheduling_class.id = test.scheduling_class
    WHERE
        task.id = NEW.task;

    IF NOT FOUND THEN
        RAISE EXCEPTION 'No task % exists.', NEW.task;
    END IF;


    IF TG_OP = 'INSERT' THEN
        NEW.added := now();
    ELSIF TG_OP = 'UPDATE' AND NEW.added <> OLD.added THEN
        RAISE EXCEPTION 'Insertion time cannot be updated.';
    END IF;



    -- Non-background gets bounced if trying to schedule beyond the
    -- scheduling horizon.

    SELECT INTO horizon schedule_horizon FROM configurables;
    IF taskrec.scheduling_class <> scheduling_class_background_multi()
       AND (lower(NEW.times) - normalized_now()) > horizon THEN
        RAISE EXCEPTION 'Cannot schedule runs more than % in advance (% is % outside the range %)',
            horizon, NEW.times, (lower(NEW.times) - normalized_now() - horizon),
	    tstzrange(normalized_now(), normalized_now()+horizon);
    END IF;

    -- Only allow time changes that shorten the run
    IF (TG_OP = 'UPDATE')
        AND ( (lower(NEW.times) <> lower(OLD.times))
              OR ( upper(NEW.times) > upper(OLD.times) ) )
    THEN
        RAISE EXCEPTION 'Runs cannot be rescheduled, only shortened.';
    END IF;

    -- Make sure UUID assignment follows a sane pattern.

    IF (TG_OP = 'INSERT') THEN

        IF taskrec.participant = 0 THEN
	    -- Lead participant should be assigning a UUID
            IF NEW.uuid IS NOT NULL THEN
                RAISE EXCEPTION 'Lead participant should not be given a run UUID.';
            END IF;
            NEW.uuid := gen_random_uuid();
        ELSE
            -- Non-leads should be given a UUID.
            IF NEW.uuid IS NULL THEN
                RAISE EXCEPTION 'Non-lead participant should not be assigning a run UUID.';
            END IF;
        END IF;

    ELSEIF (TG_OP = 'UPDATE') THEN

        IF NEW.uuid <> OLD.uuid THEN
            RAISE EXCEPTION 'UUID cannot be changed';
        END IF;

        IF NEW.state <> OLD.state AND NEW.state = run_state_canceled() THEN
	    PERFORM pg_notify('run_canceled', NEW.id::TEXT);
        END IF;

        IF NEW.state <> OLD.state AND NEW.state = run_state_running() THEN
	    UPDATE task
	    SET runs_started = runs_started + 1
	    WHERE id = NEW.task;
        END IF;

	-- TODO: Make sure part_data_full, result_ful and
	-- result_merged happen in the right order.

	NOTIFY run_change;

    END IF;


    -- TODO: When there's resource management, assign the resources to this run.

    SELECT INTO tool_name name FROM tool WHERE id = taskrec.tool; 

    -- Finished runs are what get inserted for background tasks.
    IF TG_OP = 'INSERT' AND NOT run_state_is_finished(NEW.state) THEN

        pdata_out := row_to_json(t) FROM ( SELECT taskrec.participant AS participant,
                                           cast ( taskrec.json #>> '{test}' AS json ) AS test ) t;

        run_result := pscheduler_plugin_invoke('tool', tool_name, 'participant-data', pdata_out::TEXT );
        IF run_result.status <> 0 THEN
	    RAISE EXCEPTION 'Unable to get participant data: %', run_result.stderr;
	END IF;
        NEW.part_data := regexp_replace(run_result.stdout, '\s+$', '')::JSONB;

    END IF;

    IF (TG_OP = 'UPDATE') THEN
               
	IF NEW.priority <> OLD.priority THEN
	    RAISE EXCEPTION 'Priority cannot be changed after scheduling.';
	END IF;

	-- Runs that are canceled stay that way.
	IF NEW.state <> OLD.state AND OLD.state = run_state_canceled() THEN
	    NEW.state := OLD.state;
	END IF;

	IF NOT run_state_transition_is_valid(OLD.state, NEW.state) THEN
            RAISE EXCEPTION 'Invalid transition between states (% to %).',
                OLD.state, NEW.state;
        END IF;


        -- No status changes on future runs unless the new state is a
        -- finished one.

        IF NEW.status IS NOT NULL
           AND ( (OLD.status IS NULL) OR (NEW.status <> OLD.status) )
	   AND ( NOT run_state_is_finished(NEW.state) )
           AND lower(NEW.times) > normalized_now()
        THEN
            RAISE EXCEPTION 'Cannot set status on future runs. % / %', lower(NEW.times), normalized_now();
        END IF;


	-- Handle times for runs reaching a state where they may have
	-- been running to one where they've stopped.

	IF NEW.state <> OLD.state AND run_state_is_finished(NEW.state)
        THEN

	    -- Adjust the end times only if there's a sane case for
	    -- doing so.  If the clock is out of whack, the current
	    -- time could be less than the start time, which would
	    -- make for an invalid range.

	    IF normalized_now() >= lower(OLD.times)
            THEN
	        -- Record the actual times the run ran
	    	NEW.times_actual = tstzrange(lower(OLD.times), normalized_now(), '[]');

	    	-- If the run took less than the scheduled time, return
	    	-- the remainder to the timeline.
	    	IF upper(OLD.times) > normalized_now() THEN
	           NEW.times = tstzrange(lower(OLD.times), normalized_now(), '[]');
	    	END IF;
            END IF;

        END IF;

	-- If there's now a merged result, notify anyone watching for those.

       IF OLD.result_merged IS NULL AND NEW.result_merged IS NOT NULL
       THEN
	      PERFORM pg_notify('result_available', NEW.id::TEXT);
       END IF;


    ELSIF (TG_OP = 'INSERT') THEN

        -- Make a note that this run was put on the schedule and
        -- update the start time if we don't have one.

        UPDATE task t
        SET
            runs = runs + 1,
            -- TODO: This skews future runs when the first run slips.
            first_start = coalesce(t.first_start, t.start, lower(NEW.times))
        WHERE t.id = taskrec.id;


        -- Reject new runs that overlap with anything that isn't a
        -- finished run or where this insert would cause a conflict.
        -- This is done as the absolute last step because the entire
        -- table has to be locked.  We want that to happen for as
        -- little time as possible and only if there's potential for
        -- the run to have conflicts.  The table update above will be
        -- rolled back if this bombs out.

        IF (NOT taskrec.anytime)                       -- Test can have conflicts
           AND (NOT run_state_is_finished(NEW.state))  -- Tasks that will run

        THEN

            IF run_has_conflicts(taskrec.id, lower(NEW.times), NEW.priority)
            THEN
               RAISE EXCEPTION '%', run_conflict_message();
            END IF;

        END IF;

        PERFORM pg_notify('run_new', NEW.uuid::TEXT);

    END IF;

    RETURN NEW;
END;
$$ LANGUAGE plpgsql;

CREATE TRIGGER run_alter BEFORE INSERT OR UPDATE ON run
       FOR EACH ROW EXECUTE PROCEDURE run_alter();




DROP TRIGGER IF EXISTS run_delete ON run CASCADE;

DO $$ BEGIN PERFORM drop_function_all('run_delete'); END $$;

CREATE OR REPLACE FUNCTION run_delete()
RETURNS TRIGGER
AS $$
DECLARE
    older_than TIMESTAMP WITH TIME ZONE;
BEGIN

    -- If this isn't the last run, we're good.
    IF EXISTS (SELECT * FROM run WHERE task = OLD.task)
    THEN
        RETURN OLD;
    END IF;


    -- Remove the task if it can be considered completed.

    SELECT INTO older_than normalized_now() - keep_runs_tasks
    FROM configurables;

    DELETE FROM task
    WHERE
        id = OLD.task
        -- The first of these will be the latest known time.
        AND COALESCE(until, start, added) < older_than
        AND (
            -- Complete based on runs
            (max_runs IS NOT NULL AND runs >= max_runs)
            -- One-shot
            OR (repeat IS NULL AND repeat_cron IS NULL)
            -- Until time has passed
            OR until < older_than
            )
    ;

    RETURN OLD;
END;
$$ LANGUAGE plpgsql;

CREATE TRIGGER run_delete AFTER DELETE ON run
       FOR EACH ROW EXECUTE PROCEDURE run_delete();





-- If a task becomes disabled, remove all future runs.

DROP TRIGGER IF EXISTS run_task_disabled ON task CASCADE;

DO $$ BEGIN PERFORM drop_function_all('run_task_disabled'); END $$;

CREATE OR REPLACE FUNCTION run_task_disabled()
RETURNS TRIGGER
AS $$
BEGIN
    IF NEW.enabled <> OLD.enabled AND NOT NEW.enabled THEN

        -- Chuck future runs
        DELETE FROM run
        WHERE
            task = NEW.id
            AND lower(times) > normalized_now();

        -- Mark anything current as canceled.
	UPDATE run SET state = run_state_canceled()
	WHERE
	    run.task = NEW.id
	    AND times @> normalized_now()
	    AND state <> run_state_nonstart();
    END IF;

    RETURN NEW;
END;
$$ LANGUAGE plpgsql;

CREATE TRIGGER run_task_disabled BEFORE UPDATE ON task
   FOR EACH ROW EXECUTE PROCEDURE run_task_disabled();


-- If the scheduling horizon changes and becomes smaller, remove runs
-- that go beyond it.

DROP TRIGGER IF EXISTS run_horizon_change ON configurables CASCADE;

DO $$ BEGIN PERFORM drop_function_all('run_horizon_change'); END $$;

CREATE OR REPLACE FUNCTION run_horizon_change()
RETURNS TRIGGER
AS $$
BEGIN

    IF NEW.schedule_horizon < OLD.schedule_horizon THEN
        DELETE FROM run
        WHERE upper(times) > (normalized_now() + NEW.schedule_horizon);
    END IF;

    RETURN NEW;
END;
$$ LANGUAGE plpgsql;

CREATE TRIGGER run_horizon_change AFTER UPDATE ON configurables
    FOR EACH ROW EXECUTE PROCEDURE run_horizon_change();



-- Determine if a run can proceed or is pre-empted by other runs

DO $$ BEGIN PERFORM drop_function_all('run_can_proceed'); END $$;

CREATE OR REPLACE FUNCTION run_can_proceed(
    run_id BIGINT
)
RETURNS BOOLEAN
AS $$
DECLARE
    runrec RECORD;
BEGIN
    SELECT INTO runrec
        run.*,
        test.scheduling_class,
        scheduling_class.anytime,
        scheduling_class.exclusive
    FROM
        run
        JOIN task ON task.id = run.task
        JOIN test ON test.id = task.test
        JOIN scheduling_class ON scheduling_class.id = test.scheduling_class
    WHERE run.id = run_id;


    IF NOT FOUND
    THEN
        RAISE EXCEPTION 'No such run.';
    END IF;

    -- Anytime tasks don't ever count, so they're good to go.                                                                       
    IF runrec.anytime
    THEN
        RETURN TRUE;
    END IF;

    RETURN NOT (
        -- Exclusive can't collide with anything                                                                                    
        ( runrec.exclusive
          AND EXISTS (SELECT * FROM run_conflictable
                      WHERE
                          times && runrec.times
                          AND COALESCE(priority, 0) >= COALESCE(runrec.priority, 0)
                          AND id <> run_id
                     )
        )
        -- Non-exclusive can't collide with exclusive          
          OR ( NOT runrec.exclusive
               AND EXISTS (SELECT * FROM run_conflictable
                           WHERE
                               exclusive
                               AND times && runrec.times
                               AND COALESCE(priority, 0) >= COALESCE(runrec.priority, 0)
                               AND id <> run_id
                    )
            )
        );

END;
$$ LANGUAGE plpgsql;



DO $$ BEGIN PERFORM drop_function_all('run_start'); END $$;



-- Get a run marked as started.  Return NULL if successful and an
-- error message if not.  This is used by the runner.

CREATE OR REPLACE FUNCTION run_start(run_id BIGINT)
RETURNS TEXT
AS $$
BEGIN

    IF control_is_paused()
    THEN
        UPDATE run SET
            state = run_state_missed(),
            status = 1,
            result = '{ "succeeded": false, "diags": "System was paused." }'
        WHERE id = run_id;
        RETURN 'System was paused at run time.';
    END IF;

    IF NOT run_can_proceed(run_id)
    THEN
        UPDATE run SET
            state = run_state_preempted(),
            status = 1,
            result = '{ "succeeded": false, "diags": "Run was preempted." }'
        WHERE id = run_id;
        RETURN 'Run was preempted.';
    END IF;

    UPDATE run SET state = run_state_running() WHERE id = run_id;
    RETURN NULL;

END;
$$ LANGUAGE plpgsql;


-- Maintenance functions

DO $$ BEGIN PERFORM drop_function_all('run_handle_stragglers'); END $$;

CREATE OR REPLACE FUNCTION run_handle_stragglers()
RETURNS VOID
AS $$
DECLARE
    straggle_time TIMESTAMP WITH TIME ZONE;
    straggle_time_bg_multi TIMESTAMP WITH TIME ZONE;
BEGIN

    -- When non-background-multi tasks are considered tardy
    SELECT INTO straggle_time
        normalized_now() - run_straggle FROM configurables;

    -- When non-background-multi tasks are considered tardy
    SELECT INTO straggle_time_bg_multi
        normalized_now() - run_straggle FROM configurables;

    -- Runs that failed to start
    UPDATE run
    SET state = run_state_missed()
    WHERE id IN (
        SELECT
            run.id
        FROM
            run
            JOIN task ON task.id = run.task
            JOIN test ON test.id = task.test
        WHERE

            -- Non-background-multi runs pending on deck after start times
            -- were missed
            ( test.scheduling_class <> scheduling_class_background_multi()
              AND lower(times) < straggle_time
              AND run.state IN ( run_state_pending(), run_state_on_deck() )
            )

            OR

            -- Background-multi runs that passed their end time
            ( test.scheduling_class = scheduling_class_background_multi()
              AND upper(times) < straggle_time_bg_multi
              AND run.state IN ( run_state_pending(), run_state_on_deck() )
            )
    );


    -- Straggling background-multi runs stuck in the on-deck state get
    -- knocked back into pending so the runner will try and start them
    -- again.  If they keep failing, the code above will eventually
    -- mark them missed.

    UPDATE run
    SET state = run_state_pending()
    WHERE id IN (
        SELECT run.id
        FROM
           run
           JOIN task ON task.id = run.task
           JOIN test ON test.id = task.test
           JOIN scheduling_class ON scheduling_class.id = test.scheduling_class
        WHERE
            run.times @> straggle_time
            AND run.state = run_state_on_deck()
            AND scheduling_class.multi_result
    );


    -- Runs that started and didn't report back in a timely manner
    UPDATE run
    SET state = run_state_overdue()
    WHERE
        upper(times) < straggle_time
        AND state = run_state_running();

END;
$$ LANGUAGE plpgsql;


DO $$ BEGIN PERFORM drop_function_all('run_remove_old_scheduled'); END $$;

CREATE OR REPLACE FUNCTION run_remove_old_scheduled()
RETURNS VOID
AS $$
BEGIN

    DELETE FROM run
    WHERE
      state = run_state_scheduling()
      AND normalized_now() > lower(times)
    ;

END;
$$ LANGUAGE plpgsql;


DO $$ BEGIN PERFORM drop_function_all('run_purge'); END $$;

CREATE OR REPLACE FUNCTION run_purge()
RETURNS VOID
AS $$
DECLARE
    purge_before TIMESTAMP WITH TIME ZONE;
BEGIN

    -- Most runs
    SELECT INTO purge_before now() - keep_runs_tasks FROM configurables;
    DELETE FROM run
    WHERE
        (
	  (upper(times) < purge_before)                       -- Runs that got a time
	  OR (upper(times) IS NULL AND added < purge_before)  -- Runs that didn't
        )
        AND run_state_is_finished(state)
    ;

    -- Extra margin for anything that might actually be running
    -- TODO: This is probably redundant.
    purge_before := purge_before - 'PT1H'::INTERVAL;
    DELETE FROM run
    WHERE
        upper(times) < purge_before
        AND state IN (SELECT id FROM run_state WHERE NOT finished);

    -- Runs with explicit expiration times
    DELETE FROM run
    WHERE
        expires IS NOT NULL
	AND expires < now();

END;
$$ LANGUAGE plpgsql;



-- Maintenance that happens four times a minute.

DO $$ BEGIN PERFORM drop_function_all('run_main_fifteen'); END $$;

CREATE OR REPLACE FUNCTION run_maint_fifteen()
RETURNS VOID
AS $$
BEGIN
    PERFORM run_handle_stragglers();
    PERFORM run_remove_old_scheduled();
    PERFORM run_purge();
END;
$$ LANGUAGE plpgsql;



-- Convenient ways to see the goings on

CREATE OR REPLACE VIEW run_status
AS
    SELECT
        run.id AS run,
	run.uuid AS run_uuid,
	task.id AS task,
	task.uuid AS task_uuid,
	test.name AS test,
	tool.name AS tool,
	run.times,
	run_state.display AS state
    FROM
        run
	JOIN run_state ON run_state.id = run.state
	JOIN task ON task.id = task
	JOIN test ON test.id = task.test
	JOIN tool ON tool.id = task.tool
    WHERE
        run.state <> run_state_pending()
	OR (run.state = run_state_pending()
            AND lower(run.times) < (now() + 'PT2M'::interval))
    ORDER BY run.times;


CREATE OR REPLACE VIEW run_status_short
AS
    SELECT run, task, times, state
    FROM  run_status
;




--
-- JSON Representation
--

DO $$ BEGIN PERFORM drop_function_all('run_json'); END $$;

-- Return a JSON record suitable for returning by the REST API

CREATE OR REPLACE FUNCTION run_json(run_id BIGINT)
RETURNS JSONB
AS $$
DECLARE
    rec RECORD;
    result JSONB;
BEGIN
    SELECT INTO rec
        run.*,
        run_state.*,
	task.*,
	archiving_json(run.id) AS archivings
    FROM
        run
        JOIN run_state ON run_state.id = run.state
        JOIN task ON task.id = run.task
    WHERE run.id = run_id;
    IF NOT FOUND
    THEN
        RAISE EXCEPTION 'No such run.';
    END IF;

    result := json_build_object(
        'start-time', timestamp_with_time_zone_to_iso8601(lower(rec.times)),
        'end-time', timestamp_with_time_zone_to_iso8601(upper(rec.times)),
        'duration', interval_to_iso8601(upper(rec.times) - lower(rec.times)),
	'participant', rec.participant,
	'participants', rec.participants,
	'participant-data', rec.part_data,
	'participant-data-full', rec.part_data_full,
	'result', rec.result,
	'result-full', rec.result_full,
	'result-merged', rec.result_merged,
	'state', rec.enum,
	'state-display', rec.display,
	'errors', rec.errors,
	-- clock-survey is conditional; see below.
	-- archivings is conditional; see below.
        'added', timestamp_with_time_zone_to_iso8601(rec.added),
	'priority', rec.priority,
	'limit-diags', rec.limit_diags
	-- task-href has to be added by the caller
	-- result-href has to be added by the caller
    );

    IF rec.clock_survey IS NOT NULL THEN
        result := jsonb_set(result, '{clock-survey}', rec.clock_survey::JSONB, TRUE);
    END IF;

    IF rec.archivings IS NOT NULL THEN
        result := jsonb_set(result, '{archivings}', rec.archivings::JSONB, TRUE);
    END IF;

    RETURN result;
END
$$ language plpgsql;




--
-- API
--

-- Put a run of a task on the schedule.

-- NOTE: This is for scheduled runs only, not background-multi results.

DO $$ BEGIN PERFORM drop_function_all('api_run_post'); END $$;

CREATE OR REPLACE FUNCTION api_run_post(
    task_uuid UUID,
    start_time TIMESTAMP WITH TIME ZONE,
    run_uuid UUID,  -- NULL to assign one
    nonstart_reason TEXT = NULL,
    proposed_priority INTEGER = NULL,
    limit_diags TEXT = NULL
)
RETURNS TABLE (
    succeeded BOOLEAN,  -- True if the post was successful
    new_uuid UUID,      -- UUID of run, NULL if post failed
    conflict BOOLEAN,   -- True of failed because of a conflict
    error TEXT          -- Error text if post failed
)
AS $$
DECLARE
    task RECORD;
    time_range TSTZRANGE;
    initial_state INTEGER;
    initial_status INTEGER;
    exception_text TEXT;
    run_priority INTEGER;
BEGIN

    SELECT INTO task * FROM task WHERE uuid = task_uuid;
    IF NOT FOUND THEN
        RETURN QUERY SELECT FALSE, NULL::UUID, FALSE,
            'Task does not exist'::TEXT;
        RETURN;
    END IF;

    IF run_uuid IS NULL AND task.participant <> 0 THEN
        RETURN QUERY SELECT FALSE, NULL::UUID, FALSE,
            'Cannot set run UUID as non-lead participant'::TEXT;
        RETURN;
    END IF;

    IF nonstart_reason IS NULL THEN
        IF task.participant = 0 THEN
            -- The scheduler will adjust this when scheduling is complete.
            initial_state := run_state_scheduling();
        ELSE
            initial_state := run_state_pending();
        END IF;
        initial_status := NULL;
    ELSE
        initial_state := run_state_nonstart();
        initial_status := 1;  -- Nonzero means failure.
    END IF;
    
    start_time := normalized_time(start_time);
    time_range := tstzrange(start_time, start_time + task.duration, '[)');

    -- Set priority according to what was requested

    IF proposed_priority is NULL
    THEN
        run_priority = task.priority;
    ELSE
        run_priority := LEAST(task.priority, run_priority);
    END IF;

    BEGIN

        WITH inserted_row AS (
            INSERT INTO run (uuid, task, times, state,
                errors, priority, limit_diags)
            VALUES (run_uuid, task.id, time_range, initial_state,
	        nonstart_reason, run_priority, limit_diags)
            RETURNING *
        ) SELECT INTO run_uuid uuid FROM inserted_row;

        RETURN QUERY SELECT TRUE, run_uuid, FALSE, ''::TEXT;

    EXCEPTION

        WHEN OTHERS THEN
            GET STACKED DIAGNOSTICS exception_text = MESSAGE_TEXT;
            IF exception_text <> run_conflict_message()
            THEN
                RAISE;  -- Re-reaise the original exception
            END IF;

            RETURN QUERY SELECT FALSE, NULL::UUID, TRUE, exception_text;

    END;

    RETURN;


END;
$$ LANGUAGE plpgsql;
--
-- Table of latest-scheduled runs for each task.  This is used to make
-- the query behind the schedule_runs_to_schedule view run faster.
--

DO $$
DECLARE
    t_name TEXT;            -- Name of the table being worked on
    t_version INTEGER;      -- Current version of the table
    t_version_old INTEGER;  -- Version of the table at the start
BEGIN

    --
    -- Preparation
    --

    t_name := 'run_latest';

    t_version := table_version_find(t_name);
    t_version_old := t_version;


    --
    -- Upgrade Blocks
    --

    -- Version 0 (nonexistant) to version 1
    IF t_version = 0
    THEN

        CREATE TABLE run_latest (

        	-- Task this run belongs to
        	task		BIGINT
        			UNIQUE
        			REFERENCES task(id)
        			ON DELETE CASCADE,

        	-- Range of times when this task will be run
        	latest		TIMESTAMP WITH TIME ZONE
        			NOT NULL
        );

        -- Populate the table from existing runs

        INSERT INTO run_latest
        SELECT run.task, max(upper(run.times))
        FROM
            run
            JOIN task ON task.id = run.task
        WHERE
	    task.repeat IS NOT NULL
	    OR task.repeat_cron IS NOT NULL
        GROUP BY run.task;

        CREATE INDEX run_latest_task_latest ON run_latest(task, latest);

	t_version := t_version + 1;

    END IF;


    -- Version 1 to version 2
    -- IF t_version = 1
    -- THEN
    --
    --    t_version := t_version + 1;
    --END IF;


    --
    -- Cleanup
    --

    PERFORM table_version_set(t_name, t_version, t_version_old);

END;
$$ LANGUAGE plpgsql;


DO $$ BEGIN PERFORM drop_function_all('run_latest_update'); END $$;

CREATE OR REPLACE FUNCTION run_latest_update()
RETURNS TRIGGER
AS $$
DECLARE
    affected RECORD;
    latest TIMESTAMP WITH TIME ZONE;
BEGIN

    affected := CASE
            WHEN TG_OP = 'DELETE' THEN OLD
            ELSE NEW
        END;

    -- We only care about tasks that repeat
    IF NOT EXISTS (
        SELECT * FROM task
	WHERE
	    id = affected.task
	    AND (repeat IS NOT NULL OR repeat_cron IS NOT NULL)
    )
    THEN
        RETURN affected;
    END IF;


    SELECT INTO latest
        max(upper(times))
        FROM run
        WHERE task = affected.task
        GROUP BY task;

    IF NOT FOUND OR latest IS NULL -- No run records left
    THEN
        DELETE FROM run_latest WHERE task = affected.task;
        RETURN affected;
    END IF;    

    INSERT INTO run_latest
    VALUES (affected.task, latest)
    ON CONFLICT (task) DO UPDATE
    SET latest = EXCLUDED.latest;

    RETURN NEW;
END;
$$ LANGUAGE plpgsql;


-- TODO: It would be nice if this didn't fire for every row during
-- mass deletes, but there doesn't appear to be a good way around it
-- because FOR EACH STATEMENT doesn't have a way to retrieve the rows
-- affected.

DROP TRIGGER IF EXISTS run_latest_update ON run;
CREATE TRIGGER run_latest_update AFTER INSERT OR UPDATE OR DELETE ON run
    FOR EACH ROW EXECUTE PROCEDURE run_latest_update();
--
-- Table of archiving records for a run
--

DO $$
DECLARE
    t_name TEXT;            -- Name of the table being worked on
    t_version INTEGER;      -- Current version of the table
    t_version_old INTEGER;  -- Version of the table at the start
BEGIN

    --
    -- Preparation
    --

    t_name := 'archiving';

    t_version := table_version_find(t_name);
    t_version_old := t_version;


    --
    -- Upgrade Blocks
    --

    -- Version 0 (nonexistant) to version 1
    IF t_version = 0
    THEN

        CREATE TABLE archiving (

        	-- Row identifier
        	id		BIGSERIAL
        			PRIMARY KEY,

        	-- Run this archiving is part of
        	run		BIGINT
        			NOT NULL
        			REFERENCES run(id)
        			ON DELETE CASCADE,

                -- Archiver to be used
        	archiver        BIGINT
        			NOT NULL
        			REFERENCES archiver(id)
        			ON DELETE CASCADE,

                -- Data for the archiver
        	archiver_data   JSONB,

        	-- How many times we've tried to archive the result
        	attempts    	INTEGER
        			DEFAULT 0
        			CHECK(attempts >= 0),

        	-- How many times we've tried to archive
        	last_attempt   	TIMESTAMP WITH TIME ZONE,

        	-- Whether or not this archiving has been completed,
        	-- successfully or not.
		-- Note:  This is renamed in a later version.
        	archived   	BOOLEAN
        			DEFAULT FALSE,

        	-- When we should try again if not successful
        	next_attempt  	TIMESTAMP WITH TIME ZONE
        			DEFAULT '-infinity'::TIMESTAMP WITH TIME ZONE,

        	-- Array of results from each attempt
        	diags	   	JSONB
                			DEFAULT '[]'::JSONB
        );

	t_version := t_version + 1;

    END IF;

    -- Version 1 to version 2
    -- Adds indexes for run and archiver to aid cascading deletes
    IF t_version = 1
    THEN
        CREATE INDEX archiving_run ON archiving(run);
        CREATE INDEX archiving_archiver ON archiving(archiver);

        t_version := t_version + 1;
    END IF;

    -- Version 2 to version 3
    -- Adds indexes to make archiving_eligible go faster
    IF t_version = 2
    THEN
        CREATE INDEX archiving_candidates
        ON archiving(archived, next_attempt)
	WHERE NOT archived;

        t_version := t_version + 1;
    END IF;

    -- Version 3 to version 4
    -- Improvement to archiving_candidates index
    IF t_version = 3
    THEN
        DROP INDEX IF EXISTS archiving_candidates;
        CREATE INDEX archiving_candidates
        ON archiving(archived, next_attempt)
        WHERE NOT archived AND next_attempt IS NOT NULL;

        t_version := t_version + 1;
    END IF;

    -- Version 4 to version 5
    -- Adds expiration time
    IF t_version = 4
    THEN
        -- Calculated by archiving_run_afer from the 'ttl' value in the JSON
        ALTER TABLE archiving ADD COLUMN
        ttl_expires TIMESTAMP WITH TIME ZONE;

        t_version := t_version + 1;
    END IF;


    -- Version 5 to version 6
    -- Adds JQ transformation script
    IF t_version = 5
    THEN
        ALTER TABLE archiving ADD COLUMN
        transform JSON;

        t_version := t_version + 1;
    END IF;


    -- Version 6 to version 7
    -- Make the 'archived' column more sane and add success/failure flag
    IF t_version = 6
    THEN

        ALTER TABLE task DISABLE TRIGGER USER;

        ALTER TABLE archiving RENAME COLUMN
	archived to completed;

	ALTER TABLE archiving ADD COLUMN
	archived BOOLEAN DEFAULT FALSE;

	UPDATE archiving SET archived = completed;

        ALTER TABLE task ENABLE TRIGGER USER;

        t_version := t_version + 1;
    END IF;


    -- Version 7 to version 8
    -- Added full archive spec
    IF t_version = 7
    THEN

        ALTER TABLE archiving ADD COLUMN spec JSON;

        -- Make existing rows look like they should before
        -- constraining the column.

        UPDATE archiving SET spec = json_build_object(
            'archiver', (SELECT name FROM archiver WHERE id = archiving.archiver),
            'data', archiver_data
            );

        -- Add the constraint
        ALTER TABLE archiving ALTER COLUMN spec SET NOT NULL;

        t_version := t_version + 1;
    END IF;


    --
    -- Cleanup
    --

    PERFORM table_version_set(t_name, t_version, t_version_old);

END;
$$ LANGUAGE plpgsql;



DROP TRIGGER IF EXISTS archiving_run_after ON archiving CASCADE;

DO $$ BEGIN PERFORM drop_function_all('archiving_run_after'); END $$;

CREATE OR REPLACE FUNCTION archiving_run_after()
RETURNS TRIGGER
AS $$
DECLARE
    inserted BOOLEAN;
    archive JSONB;
    expires TIMESTAMP WITH TIME ZONE;
BEGIN

    -- Start an archive if conditions are right

    IF ( -- Non-Background-multi run
         TG_OP = 'UPDATE'
         AND OLD.result_merged IS NULL 
         AND NEW.result_merged IS NOT NULL )
       OR ( -- Background-multi Run
            TG_OP = 'INSERT'
            AND NEW.state = run_state_finished()
            AND NEW.result_merged IS NOT NULL
            AND EXISTS (SELECT *
                        FROM
                            task
                            JOIN test ON test.id = task.test
                        WHERE
                            task.id = NEW.task
                            AND test.scheduling_class = scheduling_class_background_multi())
          )
   THEN

       inserted := FALSE;

        -- Insert rows into the archiving table.
        FOR archive IN (
                        -- Task-provided archivers
                        SELECT
                            jsonb_array_elements(task.json #> '{archives}')
                        FROM
                            run
                            JOIN task on run.task = task.id
                        WHERE
                            run.id = NEW.id
			    AND task.participant = 0

                        UNION ALL

                        -- System-wide default archivers
                        SELECT archive_default.archive FROM archive_default
                       )
        LOOP
            IF archive ? 'ttl'
            THEN
                expires := now() + text_to_interval(archive #>> '{ttl}');
            ELSE
                expires := NULL;
	    END IF;

	    -- If there is no "runs" value in the archive spec, treat
	    -- it as archive only on success, otherwise, see if the
	    -- final state of the run means we should do the
	    -- archiving.

	    IF NOT archive ? 'runs' THEN
	        archive := archive || '{"runs": "succeeded"}'::JSONB;
	    END IF;

	    IF archive #>> '{runs}' = 'all'
	       OR EXISTS (SELECT * FROM run_state
	                  WHERE run_state.id = NEW.state
	                  AND success = CASE
	                     WHEN archive #>> '{runs}' = 'succeeded' THEN TRUE
	                     WHEN archive #>> '{runs}' = 'failed' THEN FALSE
	                     ELSE NULL
	                     END)
	    THEN
	        INSERT INTO archiving (run, archiver, archiver_data, ttl_expires, transform, spec)
    	        VALUES (
    	            NEW.id,
    	            (SELECT id from archiver WHERE name = archive #>> '{archiver}'),
	            (archive #> '{data}')::JSONB,
	            expires,
		    (archive #> '{transform}'),
		    archive
    	        );
                inserted := TRUE;
	    END IF;
        END LOOP;

        IF inserted THEN
            NOTIFY archiving_change;
        END IF;

    END IF;

    RETURN NEW;
END;
$$ LANGUAGE plpgsql;

DROP TRIGGER IF EXISTS archiving_run_after ON run;
CREATE TRIGGER archiving_run_after AFTER INSERT OR UPDATE ON run
       FOR EACH ROW EXECUTE PROCEDURE archiving_run_after();





DROP TRIGGER IF EXISTS archiving_update ON archiving CASCADE;

DO $$ BEGIN PERFORM drop_function_all('archiving_update'); END $$;

CREATE OR REPLACE FUNCTION archiving_update()
RETURNS TRIGGER
AS $$
DECLARE
    keep_after INTERVAL;
BEGIN

    -- Notify on things the archiver will want to know about:

    IF
        -- Completions
        (NEW.completed AND NEW.completed <> OLD.completed)
        -- Reschedules
        OR (NEW.next_attempt IS NOT NULL
                AND NEW.next_attempt <> OLD.next_attempt)
    THEN

	-- Set an expiration time on the run if the task has a
	-- keep_after_archiving and this is the last-completed archiving.

	IF NOT EXISTS (SELECT * FROM archiving
	               WHERE archiving.run = NEW.run
		       AND NOT archiving.completed)
	THEN
	    SELECT INTO keep_after task.keep_after_archive
	    FROM task JOIN run ON task.id = run.task
	    WHERE run.id = NEW.run;

	    IF keep_after IS NOT NULL
	    THEN
	        UPDATE run SET expires = now() + keep_after
		WHERE id = NEW.run;
	    END IF;

	END IF;

	-- Let interested parties know what's happening.

        NOTIFY archiving_change;

    END IF;

    RETURN NEW;
END;
$$ LANGUAGE plpgsql;

DROP TRIGGER IF EXISTS archiving_update ON archiving;
CREATE TRIGGER archiving_update AFTER UPDATE ON archiving
       FOR EACH ROW EXECUTE PROCEDURE archiving_update();



-- Return the first max_return items eligible for archiving

DO $$ BEGIN PERFORM drop_function_all('archiving_next'); END $$;

CREATE OR REPLACE FUNCTION archiving_next(
    max_return INTEGER,
    exclusions BIGINT[]
)
RETURNS TABLE (
    id BIGINT,
    task_uuid UUID,
    run_uuid UUID,
    archiver TEXT,
    archiver_data JSONB,
    start TIMESTAMP WITH TIME ZONE,
    duration INTERVAL,
    test JSONB,
    tool JSONB,
    participants JSONB,
    result JSONB,
    attempts INTEGER,
    last_attempt TIMESTAMP WITH TIME ZONE,
    transform JSON,
    task_detail JSONB,
    run_detail JSONB,
    spec JSON,
    debug BOOLEAN
)
AS $$
BEGIN

    RETURN QUERY

    SELECT
        archiving.id AS id,
        task.uuid AS task_uuid,
        run.uuid AS run_uuid,
        archiver.name AS archiver,
        archiving.archiver_data,
        lower(run.times) AS start,
        task.duration AS duration,
        task.json #> '{test}' AS test,
        tool.json AS tool,
        task.participants AS participants,
        run.result_merged AS result,
        archiving.attempts AS attempts,
        archiving.last_attempt AS last_attempt,
	archiving.transform AS transform,
	-- TODO: This covers a number of things above.  Remove the
	-- redundancies here and in the archiver.
	task.json_detail AS task_detail,
	run_json(run.id) AS run_detail,
	archiving.spec AS spec,
	(task.json ->> 'debug')::BOOLEAN AS debug
    FROM
        archiving
        JOIN archiver ON archiver.id = archiving.archiver
        JOIN run ON run.id = archiving.run
        JOIN task ON task.id = run.task
        JOIN tool ON tool.id = task.tool
    WHERE
        archiving.id IN (
            SELECT archiving.id FROM archiving
            WHERE
                NOT completed
                AND next_attempt IS NOT NULL
                AND next_attempt < now()
                AND (ttl_expires IS NULL OR ttl_expires > now())
                AND archiving.id NOT IN (SELECT UNNEST(exclusions))
            ORDER BY archiving.attempts, next_attempt
            LIMIT max_return
        )
    ORDER BY attempts, next_attempt;

    RETURN;

END;
$$ LANGUAGE plpgsql;


-- A handy view for the archiver to use

-- TODO: Remove this after GA release
DROP VIEW IF EXISTS archiving_eligible;
CREATE OR REPLACE VIEW archiving_eligible
AS
    SELECT
        archiving.id,
        run.uuid,
	archiver.name AS archiver,
	archiving.archiver_data,
	lower(run.times) AS start,
	task.duration,
	task.json #> '{test}' AS test,
	tool.json AS tool,
	task.participants,
	run.result_full AS result_full,
	run.result_merged AS result,
	archiving.attempts,
	archiving.last_attempt
    FROM
        archiving
	JOIN archiver ON archiver.id = archiving.archiver
	JOIN run ON run.id = archiving.run
	JOIN task ON task.id = run.task
	JOIN tool ON tool.id = task.tool
    WHERE
        NOT completed
        AND next_attempt < now()
    ORDER BY next_attempt
;



-- Pull a run's archiving information as JSON

DO $$ BEGIN PERFORM drop_function_all('archiving_json'); END $$;

CREATE OR REPLACE FUNCTION archiving_json(run_id BIGINT)
RETURNS JSON
AS $$
DECLARE
    json_result JSON;
BEGIN

    SELECT INTO json_result
        array_to_json(array_agg(row_to_json(t)))
    FROM (
        SELECT
            archiving.spec AS spec,
            archiver.json AS archiver,
            archiving.archiver_data AS archiver_data,
            archiving.completed AS completed,
            archiving.archived AS archived,
            archiving.last_attempt AS last_attempt,
            archiving.diags AS diags
        FROM
            archiving
            JOIN archiver ON archiver.id = archiving.archiver
        WHERE run = run_id
    ) t;

    RETURN json_result;
END;
$$ LANGUAGE plpgsql;
;



--
-- Maintenance
--

-- Maintenance functions

DO $$ BEGIN PERFORM drop_function_all('archiving_maint_minute'); END $$;

CREATE OR REPLACE FUNCTION archiving_maint_minute()
RETURNS VOID
AS $$
DECLARE
    diag JSONB;
    default_next_attempt TIMESTAMP WITH TIME ZONE;
BEGIN

    -- Force archivings that have seen no attempts by their TTL into a
    -- failed state.

    diag := '{ "return-code": 1,
               "stdout": "",
               "stderr": "Time to live expired"}';
    diag := jsonb_set(diag, '{"time"}',
        to_jsonb(timestamp_with_time_zone_to_iso8601(normalized_now())), TRUE);

    UPDATE archiving
    SET
        completed = TRUE,     -- Not really, but gets the attempt off the table.
        next_attempt = NULL,
        diags = diags || diag::JSONB
    WHERE
        NOT completed
        AND ttl_expires < now();


END;
$$ LANGUAGE plpgsql;
--
-- Functions that manage the schedule of runs
--


-- Schedule, for use by the REST API

DROP VIEW IF EXISTS schedule;
CREATE OR REPLACE VIEW schedule
AS
    SELECT
        run.times,
	run.priority,
        task.uuid AS task,
        run.uuid AS run,
        run_state.enum AS state_enum,
        run_state.display AS state_display,
	run.errors AS errors,
        -- TODO: Pull full JSON with details when that's available.  See #95.
        task.json AS task_json,
	task.cli AS task_cli,
	tool.json AS tool_json,
	test.json as test_json
    FROM
        run
        JOIN run_state ON run_state.id = run.state
        JOIN task ON task.id = run.task
	JOIN tool ON tool.id = task.tool
	JOIN test ON test.id = task.test
    ORDER BY run.times
;



-- What's coming on the schedule and what to execute

-- TODO: Can probably clean this up a bit since the runner won't be
-- using as much of it.

DROP VIEW IF EXISTS schedule_upcoming;
CREATE OR REPLACE VIEW schedule_upcoming
AS
    SELECT
      run.id AS run,
      lower(run.times) - normalized_wall_clock() AS start_in,
      tool.name AS tool,
      -- Note that these bits have to be done in pieces because PgSQL
      -- has no way to turn dates and intervals to ISO 8601.
      -- TODO: See if this is actually the case.
      task.participant AS participant,
      run.part_data_full AS part_data_full,
      task.json #> '{test}' AS test,
      lower(run.times) AS start,
      task.duration AS duration,
      test.scheduling_class
    FROM
        run
	JOIN task ON task.id = run.task
	JOIN tool ON tool.id = task.tool
        JOIN test ON test.id = task.test
    WHERE
        run.state = run_state_pending()
        AND lower(run.times) >= normalized_now()
    ORDER BY run.times;
;



-- What tasks need a run scheduled and when.  This explicitly excludes
-- background tasks, which are handled separately.

DROP VIEW IF EXISTS schedule_runs_to_schedule;
CREATE OR REPLACE VIEW schedule_runs_to_schedule
AS

    -- TODO: Need to go through this and make sure the subqueries
    -- aren't selecting columns that aren't used.

    WITH interim AS (

        -- Non-repeating tasks with no runs scheduled

        SELECT
            task.id AS task,
	    task.uuid,
	    participant_key AS key,
            task.enabled,
            task.added,
            task.duration,
            task.slip,
            task.max_runs,
            task.runs,
            task.until,
            greatest(normalized_now(), task.start) AS trynext,
	    task.participant,
            test.scheduling_class,
	    task.json,
	    task.participants,
	    task.json -> 'debug' AS debug
        FROM
            task
            JOIN test ON test.id = task.test
        WHERE
	    repeat IS NULL
	    AND repeat_cron IS NULL
	    AND NOT EXISTS (SELECT * FROM run WHERE run.task = task.id)

        UNION ALL

        -- Interval-repeating tasks without runs

        SELECT
            task.id AS task,
            task.uuid,
	    participant_key AS key,
            task.enabled,
            task.added,
            task.duration,
            task.slip,
            task.max_runs,
            task.runs,
            task.until,
            greatest(task.start, normalized_now()) AS trynext,
            task.participant,
            test.scheduling_class,
	    task.json,
	    task.participants,
	    task.json -> 'debug' AS debug
        FROM
            task
            JOIN test on test.id = task.test
        WHERE
	    task.repeat IS NOT NULL
            AND (until IS NULL OR until > normalized_now())
	    AND runs = 0

	UNION ALL

        -- Interval-repeating tasks with runs

        SELECT
            task.id AS task,
            task.uuid,
	    participant_key AS key,
            task.enabled,
            task.added,
            duration,
            task.slip,
            task.max_runs,
            task.runs,
            task.until,
            task_next_run(task.first_start,
                         greatest(normalized_now(), task.start, run_latest.latest),
                         task.repeat, NULL) AS trynext,
            task.participant,
            (SELECT scheduling_class FROM test WHERE test.id = task.test) AS scheduling_class,
	    task.json,
	    task.participants,
	    task.json -> 'debug' AS debug
       FROM
            task
	    JOIN run_latest ON run_latest.task = task.id
        WHERE
            task.repeat IS NOT NULL
	    AND task.runs > 0
	    AND EXISTS (SELECT * FROM run WHERE run.task = task.id)

	UNION ALL

        -- Cron-repeating tasks with or without runs

        SELECT
            task.id AS task,
            task.uuid,
	    participant_key AS key,
            task.enabled,
            task.added,
            duration,
            task.slip,
            task.max_runs,
            task.runs,
            task.until,
            task_next_run(task.first_start,
                         greatest(
			     normalized_now(),
			     task.start,
			     task.first_start,
			     (SELECT latest FROM run_latest where task = task.id)
			 ),
                         NULL, task.repeat_cron) AS trynext,
            task.participant,
            (SELECT scheduling_class FROM test WHERE test.id = task.test) AS scheduling_class,
	    task.json,
	    task.participants,
	    task.json -> 'debug' AS debug
       FROM
            task
        WHERE
            task.repeat_cron IS NOT NULL
	    AND task.enabled

    )
    SELECT
        task,
	uuid,
	key,
	runs,
        trynext,
	slip,
	scheduling_class.anytime,
	json,
	participants,
	debug
    FROM
        interim
        JOIN scheduling_class
             ON scheduling_class.id = interim.scheduling_class,
        configurables
    WHERE
        enabled
	AND participant = 0
        AND ( (max_runs IS NULL) OR (runs < max_runs) )
        AND ( (until IS NULL) OR (trynext < until) )
	-- This grabs Anything that fits the scheduling horizon or
	-- runs background-multi.  Note that we adjust out anything
	-- where the amount of slip could put a run's start time over
	-- the edge of the horizon.  The horizon is far enough out
	-- that once the start time falls far enough that slip will
	-- work, the scheduler will pick it up in plenty of time.  See
	-- #1064 for the bug that brought this behavior about.
        AND (
            trynext < (
	        normalized_now()                   -- From now
		+ schedule_horizon                 -- To the horizon
		- COALESCE(slip, 'P0D'::INTERVAL)  -- Less any slip that might put us over
	    )
            OR scheduling_class = scheduling_class_background_multi()
        )
    ORDER BY trynext, added
;




-- Runs that overlap but shouldn't (for diagnostic use)
DROP VIEW IF EXISTS schedule_overlap;
CREATE OR REPLACE VIEW schedule_overlap
AS
    SELECT
        r1.id AS r1_id,
        r1.task AS r1_task,
        r2.id AS r2_id,
        r2.task AS r2_task,
        r1.times AS r1_times,
        r2.times AS r2_times
    FROM
        run_conflictable r1
        JOIN run_conflictable r2
            ON r2.id <> r1.id
            AND r1.times && r2.times
            AND r1.state <> run_state_nonstart()
            AND r2.state <> run_state_nonstart()
            AND r1.added < r2.added
        ORDER BY
          r1.times,
          r2.times
;



-- Return a schedule with bounded past and future window sizes, mostly
-- for use by the monitor.

DO $$ BEGIN PERFORM drop_function_all('schedule_monitor'); END $$;

CREATE OR REPLACE FUNCTION schedule_monitor(
    window_size INTEGER
)
RETURNS TABLE (
    ppf INTEGER,
    times TSTZRANGE,
    task UUID,
    run UUID,
    state_enum TEXT,
    state_display TEXT,
    task_json JSONB,
    task_cli JSON,
    priority INTEGER
)
AS $$
DECLARE
    normalized_time TIMESTAMP WITH TIME ZONE;
BEGIN

    normalized_time := normalized_now();

    RETURN QUERY

        SELECT

            CASE  -- Past/Present/Future
                WHEN upper(run.times) < normalized_time THEN -1
                WHEN lower(run.times) > normalized_time THEN 1
                ELSE 0
            END AS ppf,
            run.times,
            task.uuid,
            run.uuid,
            run_state.enum,
            run_state.display,
            task.json,
            task.cli,
            run.priority
        FROM
            run
            JOIN task on task.id = run.task
            JOIN run_state on run_state.id = run.state
        WHERE
            run.id IN (

                SELECT * FROM (
                    SELECT id FROM run
                    WHERE upper(run.times) < normalized_time
                    ORDER BY upper(run.times) DESC
                    LIMIT window_size
                ) past

                UNION ALL

                SELECT * FROM (
                    SELECT id FROM run
                    WHERE run.times @> normalized_time
                    ORDER BY run.times
                ) present

                UNION ALL

                SELECT * FROM (
                    SELECT id FROM run
                    WHERE
                        lower(run.times) > normalized_time
                        AND run.state IN (
			    run_state_pending(),
			    run_state_on_deck(),
			    run_state_nonstart()
			    )
                    ORDER BY lower(run.times)
                    LIMIT window_size
                ) future

            )
        ORDER BY ppf, times, task.added ASC
        ;

    RETURN;

END;
$$ LANGUAGE plpgsql;



--
-- API
--


-- Come up with a list of proposed times to run a task that fall
-- within a given time range.  If the start of the range is NULL, the
-- current time will be used.  If the end is NULL, the start plus the
-- length of the configured schedule time horizon will be used.  Under
-- no circumstances will any times in the past or beyond the time
-- horizon be proposed.

DO $$ BEGIN PERFORM drop_function_all('api_proposed_times'); END $$;

CREATE OR REPLACE FUNCTION api_proposed_times(
    task_uuid UUID,
    range_start TIMESTAMP WITH TIME ZONE = normalized_now(),
    range_end TIMESTAMP WITH TIME ZONE = tstz_infinity(),
    proposed_priority INTEGER = priority_min()
)
RETURNS TABLE (
    lower TIMESTAMP WITH TIME ZONE,
    upper TIMESTAMP WITH TIME ZONE
)
AS $$
DECLARE
    time_now TIMESTAMP WITH TIME ZONE;
    horizon_end TIMESTAMP WITH TIME ZONE;
    taskrec RECORD;
    time_range TSTZRANGE;
    last_end TIMESTAMP WITH TIME ZONE;
    run_record RECORD;
    use_priority INTEGER;
BEGIN

    -- Validate the input

    SELECT INTO taskrec
        task.*,
	test.scheduling_class,
        scheduling_class.anytime,
        scheduling_class.exclusive
    FROM
        task
        JOIN test ON test.id = task.test
        JOIN scheduling_class ON scheduling_class.id = test.scheduling_class
    WHERE
        uuid = task_uuid;

    IF NOT FOUND THEN
        RAISE EXCEPTION 'No task % exists', task_uuid;
    END IF;

    IF range_end < range_start THEN
        RAISE EXCEPTION 'Range start and end must be in the proper order';
    END IF;


    -- If the scheduling class for the task is background, the entire
    -- range is fair game and that's our final answer.
    IF taskrec.anytime THEN
        RETURN QUERY
            SELECT range_start AS lower, range_end AS upper;
        RETURN;
    END IF;


    -- Figure out the actual boundaries of where we can schedule and
    -- trim accordingly.

    time_now := normalized_now();
    SELECT INTO horizon_end time_now + schedule_horizon FROM configurables;

    -- Pass on ranges for which nothing can be scheduled
    IF 
        range_end <= time_now         -- Completely in the past
        OR range_start > horizon_end  -- Too far in the future
    THEN
        RETURN;
    END IF;


    -- Correct for the range start being earlier than now.  Don't
    -- bother proposing anything for a run_start_margin's worth of
    -- time, either.

    -- TODO: This might be more precicely accomplished by checking for
    -- tasks that would collide, but scheduling something so close to
    -- the present is relatively unlikely.

    range_start := greatest(range_start,
    		   	    time_now + (SELECT run_start_margin FROM configurables));
    range_start := normalized_time(range_start);

    -- Can't schedule past the end of the time horizon, either.
    range_end := LEAST(range_end, horizon_end);
    range_end := normalized_time(range_end);

    -- If the adjusted start overshoots the end, there's no time
    -- available.  Punt.
    IF range_start >= range_end THEN
        RETURN;
    END IF;

    time_range := tstzrange(range_start, range_end, '[)');

    last_end := range_start;


    -- Figure out what priority will be used.

    -- TODO: This and code in run.sql (~line 975) are almost close
    -- enough to merit writing a function to do this.

    use_priority := proposed_priority;
    IF use_priority IS NOT NULL
    THEN
        IF taskrec.priority IS NOT NULL
	THEN
	    -- No priority for the task means no priority for the run.
	    use_priority := NULL;
	ELSE	
	    -- Don't exceed the prioirity assigned to the task.
	    use_priority := LEAST(use_priority, taskrec.priority);
	END IF;
    END IF;


    -- Sift through everything on the timeline that overlaps with the
    -- time range and find the gaps.  Other than non-starters, the
    -- runs we care about avoiding are represented by this truth
    -- table:
    --
    -- Proposed  ||         Run on Timeline         |
    -- Run       || Background | Normal | Exclusive |
    -- ----------++------------+--------+-----------+
    -- Background|| Ignore     | Ignore |   Ignore  |
    -- Normal    || Ignore     | Ignore |   Avoid   |
    -- Exclusive || Ignore     | Avoid  |   Avoid   |

    -- TODO: Can this be selected out of run_conflictable?

    FOR run_record IN
        SELECT run.*
        FROM
            run
            JOIN task ON task.id = run.task
            JOIN test ON test.id = task.test
            JOIN scheduling_class
                 ON scheduling_class.id = test.scheduling_class
        WHERE
            -- Overlap
	    times && time_range
	    -- Higher priority than proposed or already running
	    AND ( run.priority >= use_priority
	          OR run.state = run_state_running() )
            -- Ignore non-starters
            AND state <> run_state_nonstart()
            -- Ignore background
            AND NOT scheduling_class.anytime
            AND (
                -- Always avoid exclusive runs
                (scheduling_class.exclusive)
                -- Avoid normal runs if the proposed run is exclusive
                OR (taskrec.exclusive AND NOT scheduling_class.exclusive)
            )
        ORDER BY times
    LOOP

        -- Clamp the end time to the horizon
        IF upper(run_record.times) > horizon_end
        THEN
            run_record.times = tstzrange( lower(run_record.times), horizon_end,
                '[)' );
        END IF;

        -- A run that starts beyond the last end implies a gap, but
        -- the gap must be longer than the duration of the task for it
        -- to one that can be proposed.
        IF lower(run_record.times) > last_end
           AND lower(run_record.times) - last_end >= taskrec.duration
        THEN
            RETURN QUERY
                SELECT last_end AS lower, lower(run_record.times) AS upper;
        END IF;

        last_end := upper(run_record.times);

    END LOOP;

    -- If the end of the last run is before the end of the range,
    -- that's also a gap.
    IF last_end < range_end
       AND range_end - last_end >= taskrec.duration
    THEN
        RETURN QUERY SELECT last_end, range_end;
    END IF;

    RETURN;

END;
$$ LANGUAGE plpgsql;



--
-- Maintenance
--

DO $$ BEGIN PERFORM drop_function_all('schedule_maint_minute'); END $$;

CREATE OR REPLACE FUNCTION schedule_maint_minute()
RETURNS VOID
AS $$
DECLARE
    older_than TIMESTAMP WITH TIME ZONE;
BEGIN

    SELECT INTO older_than normalized_now() - keep_runs_tasks
    FROM configurables;

    -- Get rid of tasks that no longer have runs and can be considered
    -- completed.  Note this this is done here and not task.sql
    -- because it depends on that table and run.

    DELETE FROM task
    WHERE
        NOT EXISTS (SELECT * FROM run where run.task = task.id)
        -- Use time added as a proxy for start time in non-repeaters
        AND COALESCE(start, added) < older_than
        AND (
            -- Complete based on runs
            (max_runs IS NOT NULL AND runs >= max_runs)
            -- One-shot
            OR repeat IS NULL
            )
    ;

END;
$$ LANGUAGE plpgsql;
--
-- Functions for handling various kinds of boots
--

DO $$ BEGIN PERFORM drop_function_all('warm_boot'); END $$;

CREATE OR REPLACE FUNCTION warm_boot()
RETURNS VOID
AS $$
BEGIN
    -- These are nondestructive and load the lists.
    PERFORM test_boot();
    PERFORM tool_boot();
    PERFORM archiver_boot();
    PERFORM context_boot();
    NOTIFY warmboot;
END;
$$ LANGUAGE plpgsql;


-- This is invoked by the ticker when it starts

DO $$ BEGIN PERFORM drop_function_all('cold_boot'); END $$;

CREATE OR REPLACE FUNCTION cold_boot()
RETURNS VOID
AS $$
BEGIN
    PERFORM heartbeat_cold_boot();
    PERFORM warm_boot();
END;
$$ LANGUAGE plpgsql;
--
-- Ticker Stored Procedure
--


-- Things that get done a fifteen-second intervals

DO $$ BEGIN PERFORM drop_function_all('ticker_fifteen'); END $$;

CREATE OR REPLACE FUNCTION ticker_fifteen()
RETURNS VOID
AS $$
BEGIN
    PERFORM task_maint_fifteen();
    PERFORM run_maint_fifteen();
END;
$$ LANGUAGE plpgsql;


-- Things that get done a one-minute intervals

DO $$ BEGIN PERFORM drop_function_all('ticker_minunte'); END $$;

CREATE OR REPLACE FUNCTION ticker_minute()
RETURNS VOID
AS $$
BEGIN
    PERFORM schedule_maint_minute();
    PERFORM archiving_maint_minute();
    PERFORM configurables_maint_minute();
END;
$$ LANGUAGE plpgsql;


-- Things that get done on the hour

DO $$ BEGIN PERFORM drop_function_all('ticker_hour'); END $$;

CREATE OR REPLACE FUNCTION ticker_hour()
RETURNS VOID
AS $$
BEGIN
    NULL;
END;
$$ LANGUAGE plpgsql;


-- Things that get done daily, but not necessarily at midnight

DO $$ BEGIN PERFORM drop_function_all('ticker_day'); END $$;

CREATE OR REPLACE FUNCTION ticker_day()
RETURNS VOID
AS $$
BEGIN
    NULL;
END;
$$ LANGUAGE plpgsql;


-- Things that get done at midnight

DO $$ BEGIN PERFORM drop_function_all('ticker_midnight'); END $$;

CREATE OR REPLACE FUNCTION ticker_midnight()
RETURNS VOID
AS $$
BEGIN
    NULL;
END;
$$ LANGUAGE plpgsql;



-- Cron on a budget.  This should be called once per minute,
-- preferably as close to the minute boundary as possible.  This
-- should be called repeatedly from outside to prevent having a
-- long-running transaction.

DO $$ BEGIN PERFORM drop_function_all('ticker'); END $$;

CREATE OR REPLACE FUNCTION ticker()
RETURNS INTERVAL
AS $$
DECLARE
    this_minute TIMESTAMP WITH TIME ZONE;
    clock_time TIMESTAMP WITH TIME ZONE;
BEGIN

    PERFORM ticker_fifteen();

    -- Allow a small amount of slop for minute intervals.
    IF EXTRACT(SECOND FROM now()) BETWEEN 0 and 5 THEN
        PERFORM ticker_minute();
    END IF;

    this_minute := date_trunc('minute', now());

    IF this_minute = date_trunc('hour', now()) THEN
        PERFORM ticker_hour();
    END IF;

    IF this_minute = date_trunc('day', now()) THEN
        PERFORM ticker_midnight();
    END IF;

    -- Use the actual wall clock time to determine how long until the
    -- next call (0/15/30/45 seconds within each minute).  This will
    -- compensate for any time consumed running the periodic
    -- functions.

    clock_time := clock_timestamp();  -- Note:  Not standard SQL
    RETURN
        date_trunc('minute', clock_time)
        + (TRUNC(EXTRACT(SECONDS FROM clock_time) / 15) + 1)
           * 'PT15S'::interval
        - (clock_time - 'PT1S'::interval);

END;
$$ LANGUAGE plpgsql;
--
-- Postamble for non-privileged portion of database
--

COMMIT;
